"""This module contains the definition of the ActivePlasmaLens class"""

from typing import Optional, Union, Callable, Literal

import numpy as np
import scipy.constants as ct

from .plasma_stage import PlasmaStage, DtBunchType
from wake_t.physics_models.em_fields.linear_b_theta import LinearBThetaField


class ActivePlasmaLens(PlasmaStage):
    """
    Class defining an active plasma lens.

    This elements is a subclass of :class:`PlasmaStage`, where a linear
    azimuthal magnetic field is added externally. It also includes
    convenient methods to specify the field gradient and whether the plasma
    wakefields should be taken into account.

    Parameters
    ----------
    length : float
        Length of the plasma lens in :math:`m`.
    foc_strength : float
        Focusing strength of the plasma lens in :math:`T/m`. Defined so that
        a positive value is focusing for electrons.
    wakefields : bool
        If ``True``, the beam-induced wakefields in the plasma lens will be
        computed using the model specified in ``'wakefield_model'`` and
        taken into account for the beam evolution.
    wakefield_model : str
        Name of the model which should be used for computing the
        beam-induced wakefields. Recommended models are ``'cold_fluid_1d'`` or
        ``'quasistatic_2d'``. See :class:`PlasmaStage` documentation for other
        possibilities.
    density : float or callable
        Optional. Required only if ``wakefields=True``. Plasma density
        of the APL in units of :math:`m^{-3}`. See :class:`PlasmaStage`
        documentation for more details.
    bunch_pusher : str
        The pusher used to evolve the particle bunches in time within
        the specified fields. Possible values are ``'rk4'`` (Runge-Kutta
        method of 4th order) or ``'boris'`` (Boris method).
    dt_bunch : float
        The time step for evolving the particle bunches. If ``None``, it will
        be automatically set to :math:`dt = T/(10*2*pi)`, where T is the
        smallest expected betatron period of the bunch along the plasma lens
        (T is calculated from `foc_strength` if `wakefields=False`,
        otherwise the focusing strength of a blowout is used).
        A list of values can also be provided. In this case, the list
        should have the same order as the list of bunches given to the
        ``track`` method.
    push_bunches_before_diags : bool, optional
        Whether to push the bunches before saving them to the diagnostics.
        Since the time step of the diagnostics can be different from that
        of the bunches, it could happen that the bunches appear in the
        diagnostics as they were at the last push, but not at the actual
        time of the diagnostics. Setting this parameter to ``True``
        (default) ensures that an additional push is given to all bunches
        to evolve them to the diagnostics time before saving.
        This additional push will always have a time step smaller than
        the the time step of the bunch, so it has no detrimental impact
        on the accuracy of the simulation. However, it could make
        convergence studies more difficult to interpret,
        since the number of pushes will depend on `n_diags`. Therefore,
        it is exposed as an option so that it can be disabled if needed.
    n_out : int
        Number of times along the lens in which the particle distribution
        should be returned (A list with all output bunches is returned
        after tracking).
    name : str
        Name of the plasma lens. This is only used for displaying the
        progress bar during tracking. By default, ``'Active plasma lens'``.
    **model_params
        Optional. Required only if ``wakefields=True``. Keyword arguments
        which will be given to the wakefield model. See :class:`PlasmaStage`
        documentation for more details.

    See Also
    --------
    PlasmaStage

    """

    def __init__(
        self,
        length: float,
        foc_strength: float,
        wakefields: bool = False,
        density: Optional[Union[float, Callable[[float], float]]] = None,
        wakefield_model: Optional[str] = "quasistatic_2d",
        bunch_pusher: Optional[Literal["boris", "rk4"]] = "boris",
        dt_bunch: Optional[DtBunchType] = "auto",
        push_bunches_before_diags: Optional[bool] = True,
        n_out: Optional[int] = 1,
        name: Optional[str] = "Active plasma lens",
        **model_params,
    ) -> None:
        self.foc_strength = foc_strength
        self.wakefields = wakefields
        if not self.wakefields:
            wakefield_model = None
        if density is None:
            if wakefields:
                raise ValueError(
                    "A density value is required to compute plasma wakefields"
                )
            else:
                # Give any value (it won't be used.)
                density = 0.0
        self.apl_field = LinearBThetaField(-self.foc_strength)
        super().__init__(
            length=length,
            density=density,
            wakefield_model=wakefield_model,
            bunch_pusher=bunch_pusher,
            dt_bunch=dt_bunch,
            push_bunches_before_diags=push_bunches_before_diags,
            n_out=n_out,
            name=name,
            external_fields=[self.apl_field],
            **model_params,
        )

    def _get_optimized_dt(self, beam):
        """Get tracking time step."""
        # If plasma wakefields are active, use default dt.
        if self.wakefields:
            dt = super()._get_optimized_dt(beam)
        # Otherwise, determine dt from the APL focusing strength.
        else:
            # Get minimum gamma in the bunch (assumes px,py << pz).
            q_over_m = beam.q_species / beam.m_species
            min_gamma = np.sqrt(np.min(beam.pz) ** 2 + 1)
            w_x = np.sqrt(np.abs(q_over_m * ct.c * self.foc_strength / min_gamma))
            T_x = 1 / w_x
            dt = 0.1 * T_x
        return dt



"""
Contains the method to compute the azimuthal magnetic field from the plasma
according to the paper by P. Baxevanis and G. Stupakov.

"""

from wake_t.utilities.numba import njit_serial


@njit_serial()
def calculate_b_theta_at_particles(
    r_e,
    pr_e,
    w_e,
    w_center_e,
    gamma_e,
    q_e,
    r_i,
    ion_motion,
    psi_e,
    dr_psi_e,
    dxi_psi_e,
    b_t_0_e,
    nabla_a2_e,
    A,
    B,
    C,
    K,
    U,
    a_0,
    a,
    b,
    b_t_e,
    b_t_i,
):
    """Calculate the azimuthal magnetic field at the plasma particles.

    To simplify the algorithm, this method considers only the magnetic field
    generated by the electrons, not the ions. This is usually a reasonable
    approximation, even when ion motion is enabled, because the ions are much
    slower than the electrons.

    The value of b_theta at a a radial position r is calculated as

        b_theta = a_i * r + b_i / r

    This requires determining the values of the a_i and b_i coefficients,
    which are calculated as follows, using Eqs. (26) and (27) from the paper
    of P. Baxevanis and G. Stupakov:

        Write a_i and b_i as linear system of a_0:

            a_i = K_i * a_0_diff + T_i
            b_i = U_i * a_0_diff + P_i


        Where (im1 stands for subindex i-1):

            K_i = (1 + A_i*r_i/2) * K_im1  +  A_i/(2*r_i)     * U_im1
            U_i = (-A_i*r_i**3/2) * K_im1  +  (1 - A_i*r_i/2) * U_im1

            T_i = ( (1 + A_i*r_i/2) * T_im1  +  A_i/(2*r_i)     * P_im1  +
                    (2*Bi + Ai*Ci)/4 )
            P_i = ( (-A_i*r_i**3/2) * T_im1  +  (1 - A_i*r_i/2) * P_im1  +
                    r_i*(4*Ci - 2*Bi*r_i - Ai*Ci*r_i)/4 )

        With initial conditions:

            K_0 = 1
            U_0 = 0
            T_0 = 0
            P_0 = 0

        Then a_0 can be determined by imposing a_N = 0:

            a_N = K_N * a_0_diff + T_N = 0 <=> a_0_diff = - T_N / K_N

        If the precision of a_i and b_i becomes too low, then T_i and P_i are
        recalculated with an initial guess equal to a_i and b_i, as well as a
        new a_0_diff.

    Parameters
    ----------
    r_e, pr_e, w_e, w_center_e, gamma_e : ndarray
        Radial position, momentum, weight and Lorenz factor of the plasma
        electrons.
    q_e : float
        Charge of the plasma electron species.
    r_i : ndarray
        Radial position of the plasma ions.
    i_sort_e, i_sort_i : ndarray
        Sorted indices of the electrons and ions (from lower to higher radii).
    ion_motion : bool
        Whether the ions can move. If `True`, the magnetic field
        will also be calculated at the ions.
    r_neighbor_e : ndarray
        The radial location of the middle points between
        each electron and its left and right neighbors. This array is already
        sorted and should not be indexed with `i_sort_e`.
    psi_e, dr_psi_e, dxi_psi_e : ndarray
        Value of psi and its derivatives at the location of the plasma
        electrons.
    b_t_0_e, nabla_a2_e : ndarray
        Value of the source terms (magnetic field from bunches and
        ponderomotive force of a laser) at the location of the plasma
        electrons.
    A, B, C : ndarray
        Arrays where the A_i, B_i, C_i terms in Eq. (26) will be stored.
    K, U : ndarray
        Auxiliary arrays where terms for solving the system in Eq. (27) will
        be stored.
    a_0, a, b : ndarray
        Arrays where the a_i and b_i coefficients coming out of solving the
        system in Eq. (27) will be stored.
    b_t_e, b_t_i : ndarray
        Arrays where azimuthal magnetic field at the plasma electrons and ions
        will be stored.
    """
    # Only the magnetic field from the electrons is computed, so the equations
    # below assume that q_i/m_i = 1.

    # Calculate the A_i, B_i, C_i coefficients in Eq. (26).
    calculate_ABC(
        r_e,
        pr_e,
        gamma_e,
        psi_e,
        dr_psi_e,
        dxi_psi_e,
        b_t_0_e,
        nabla_a2_e,
        A,
        B,
        C,
    )

    # Calculate the a_i, b_i coefficients in Eq. (27).
    calculate_KU(r_e, q_e, w_e, w_center_e, A, K, U)
    calculate_ai_bi_from_axis(r_e, q_e, w_e, w_center_e, A, B, C, K, U, a_0, a, b)

    # Calculate b_theta at plasma particles.
    calculate_b_theta_at_particle_centers(a, b, r_e, b_t_e)
    check_b_theta(b_t_e)
    if ion_motion:
        calculate_b_theta_with_interpolation(r_i, a_0[0], a, b, r_e, b_t_i)
        check_b_theta(b_t_i)


@njit_serial(error_model="numpy")
def calculate_b_theta_with_interpolation(r_fld, a_0, a, b, r, b_theta):
    """
    Calculate the azimuthal magnetic field from the plasma at the radial
    locations in `r_fld`.

    """
    # Calculate fields at r_fld
    n_part = r.shape[0]
    n_points = r_fld.shape[0]
    i_last = 0
    a_i = a_0
    b_i = 0.0
    b_theta_left = 0.0
    r_left = 0.0
    for j in range(n_points):
        r_j = r_fld[j]
        # Get index of last plasma particle with r_i < r_j, continuing from
        # last particle found in previous iteration.
        while i_last < n_part:
            r_i = r[i_last]
            if r_i >= r_j:
                break
            i_last += 1
        if i_last < n_part:
            if i_last > 0:
                r_left = r[i_last - 1]
                a_i = a[i_last - 1]
                b_i = b[i_last - 1]
                b_theta_left = a_i * r_left + b_i / r_left
            r_right = r[i_last]
            a_i = a[i_last]
            b_i = b[i_last]
            b_theta_right = a_i * r_right + b_i / r_right
            slope = (b_theta_right - b_theta_left) / (r_right - r_left)
            b_theta_j = b_theta_left + slope * (r_j - r_left)
        else:
            b_theta_j = a_i * r_j + b_i / r_j
        b_theta[j] = b_theta_j


@njit_serial(error_model="numpy")
def calculate_b_theta_at_particle_centers(a, b, r, b_theta):
    """
    Calculate the azimuthal magnetic field from the plasma at the radial
    locations in `r_fld`.

    """
    # Calculate fields at r_fld
    n_part = r.shape[0]
    for i in range(n_part):
        a_i = a[i]
        b_i = b[i]
        r_i = r[i]
        b_theta[i] = a_i * r_i + b_i / r_i


@njit_serial(error_model="numpy")
def calculate_ai_bi_from_axis(r, q, w, w_center, A, B, C, K, U, a_0, a, b):
    """
    Calculate the values of a_i and b_i which are needed to determine
    b_theta at any r position.

    For details about the input parameters and algorithm see method
    'calculate_b_theta_at_particles'.

    """
    n_part = r.shape[0]

    # Establish initial conditions (T_0 = 0, P_0 = 0)
    T_im1 = 0.0
    P_im1 = 0.0

    a_0[:] = 0.0

    i_start = 0

    while i_start < n_part:
        # Iterate over particles
        for i in range(i_start, n_part):
            r_i = r[i]
            q_i = q * w[i]
            q_center_i = q * w_center[i]
            A_i = A[i]
            B_i = B[i]
            C_i = C[i]
            A_inv_r_i = A_i / r_i
            A_r_i = A_i * r_i
            A_r_i_3 = A_r_i * r_i * r_i

            # Calculate value of coefficients at the center of the particles.
            l_i = 1.0 + 0.5 * q_center_i * A_r_i
            m_i = 0.5 * q_center_i * A_inv_r_i
            n_i = -0.5 * q_center_i * A_r_i_3
            o_i = 1.0 - 0.5 * q_center_i * A_r_i
            a[i] = (
                l_i * T_im1
                + m_i * P_im1
                + 0.5 * q_center_i * B_i
                + 0.25 * q_center_i * q_center_i * A_i * C_i
            )
            b[i] = (
                n_i * T_im1
                + o_i * P_im1
                + r_i
                * (
                    q_center_i * C_i
                    - 0.5 * q_center_i * B_i * r_i
                    - 0.25 * q_center_i * q_center_i * A_i * C_i * r_i
                )
            )

            # But add total charge for next iteration.
            l_i = 1.0 + 0.5 * q_i * A_r_i
            m_i = 0.5 * q_i * A_inv_r_i
            n_i = -0.5 * q_i * A_r_i_3
            o_i = 1.0 - 0.5 * q_i * A_r_i
            T_i = (
                l_i * T_im1
                + m_i * P_im1
                + 0.5 * q_i * B_i
                + 0.25 * q_i * q_i * A_i * C_i
            )
            P_i = (
                n_i * T_im1
                + o_i * P_im1
                + r_i
                * (
                    q_i * C_i
                    - 0.5 * q_i * B_i * r_i
                    - 0.25 * q_i * q_i * A_i * C_i * r_i
                )
            )
            T_im1 = T_i
            P_im1 = P_i

        # Calculate a_0_diff.
        a_0_diff = -a[i] / K[i]
        a_0 += a_0_diff

        # Calculate a_i (in T_i) and b_i (in P_i) as functions of a_0_diff.
        i_stop = n_part
        im1 = 0
        for i in range(i_start, n_part):
            T_old = a[i]
            P_old = b[i]
            K_old = K[i] * a_0_diff
            U_old = U[i] * a_0_diff

            # Test if precision is lost in the sum T_old + K_old.
            # 0.5 roughly corresponds to one lost bit of precision.
            # Also pass test if this is the first number of this iteration
            # to avoid an infinite loop or if this is the last number
            # to computer as that is zero by construction
            # Angel: if T_old + K_old (small number) is less than 10 orders
            # of magnitude smaller than T_old - K_old (big number), then we
            # have enough precision (from simulation tests).
            if (
                i == i_start
                or i == (n_part - 1)
                or abs(T_old + K_old) >= 1e-10 * abs(T_old - K_old)
                and abs(P_old + U_old) >= 1e-10 * abs(P_old - U_old)
            ):
                # Calculate a_i and b_i as functions of a_0_diff.
                # Store the result in T and P
                a[i] = T_old + K_old
                b[i] = P_old + U_old
            else:
                # If the precision is not sufficient, stop this iteration
                # and rescale T_im1 and P_im1 for the next one.
                i_stop = i
                T_im1 = a[im1]
                P_im1 = b[im1]
                break
            im1 = i

        # Start the next iteration where this one stopped
        i_start = i_stop


@njit_serial(error_model="numpy")
def calculate_ABC(r, pr, gamma, psi, dr_psi, dxi_psi, b_theta_0, nabla_a2, A, B, C):
    """Calculate the A_i, B_i and C_i coefficients of the linear system.

    The coefficients are missing the q_i * w_i term. They are multiplied by it
    in following functions.
    """
    n_part = r.shape[0]

    for i in range(n_part):
        r_i = r[i]
        pr_i = pr[i]
        gamma_i = gamma[i]
        psi_i = psi[i]
        dr_psi_i = dr_psi[i]
        dxi_psi_i = dxi_psi[i]
        b_theta_0_i = b_theta_0[i]
        nabla_a2_i = nabla_a2[i]

        a = 1.0 + psi_i
        inv_a = 1.0 / a
        inv_a2 = inv_a * inv_a
        inv_a3 = inv_a2 * inv_a
        inv_r_i = 1.0 / r_i
        b = inv_a * inv_r_i
        c = inv_a2 * inv_r_i
        pr_i2 = pr_i * pr_i

        A[i] = b
        B[i] = (
            -(gamma_i * dr_psi_i) * c
            + (pr_i2 * dr_psi_i) * inv_r_i * inv_a3
            + (pr_i * dxi_psi_i) * c
            + pr_i2 * inv_r_i * inv_r_i * inv_a2
            + b_theta_0_i * b
            + nabla_a2_i * c * 0.5
        )
        C[i] = pr_i2 * c - (gamma_i * inv_a - 1.0) * inv_r_i


@njit_serial(error_model="numpy")
def calculate_KU(r, q, w, w_center, A, K, U):
    """Calculate the K_i and U_i values of the linear system."""
    n_part = r.shape[0]

    # Establish initial conditions (K_0 = 1, U_0 = 0)
    K_im1 = 1.0
    U_im1 = 0.0

    for i in range(n_part):
        r_i = r[i]
        q_i = q * w[i]
        q_center_i = q * w_center[i]
        A_i = A[i]
        A_inv_r_i = A_i / r_i
        A_r_i = A_i * r_i
        A_r_i_3 = A_r_i * r_i * r_i

        # Calculate value of coefficients at the center of the particles.
        l_i = 1.0 + 0.5 * q_center_i * A_r_i
        m_i = 0.5 * q_center_i * A_inv_r_i
        n_i = -0.5 * q_center_i * A_r_i_3
        o_i = 1.0 - 0.5 * q_center_i * A_r_i
        K[i] = l_i * K_im1 + m_i * U_im1
        U[i] = n_i * K_im1 + o_i * U_im1

        # But add total charge for next iteration.
        l_i = 1.0 + 0.5 * q_i * A_r_i
        m_i = 0.5 * q_i * A_inv_r_i
        n_i = -0.5 * q_i * A_r_i_3
        o_i = 1.0 - 0.5 * q_i * A_r_i
        K_i = l_i * K_im1 + m_i * U_im1
        U_i = n_i * K_im1 + o_i * U_im1
        K_im1 = K_i
        U_im1 = U_i


@njit_serial()
def check_b_theta(b_theta):
    """Check that the values of b_theta are within a reasonable range

    This is used to prevent issues at the peak of a blowout wake, for example.
    """
    for i in range(b_theta.shape[0]):
        b_theta_i = b_theta[i]
        if b_theta_i < -3:
            b_theta[i] = -3
        elif b_theta_i > 3:
            b_theta[i] = 3



import numpy as np
import scipy.constants as ct
import aptools.plasma_accel.general_equations as ge

from wake_t.fields.analytical_field import AnalyticalField
from wake_t.utilities.numba import prange


class SimpleBlowoutWakefield(AnalyticalField):
    def __init__(self, n_p, laser, field_offset=0):
        """
        [n_p] = m^-3
        """
        self.density = n_p
        self.laser = laser
        self.field_offset = field_offset

        def e_x(x, y, xi, t, ex, constants):
            k = constants[0]
            for i in prange(x.shape[0]):
                ex[i] = ct.c * k * x[i]

        def e_y(x, y, xi, t, ey, constants):
            k = constants[0]
            for i in prange(x.shape[0]):
                ey[i] = ct.c * k * y[i]

        def e_z(x, y, xi, t, ez, constants):
            e_z_p = constants[1]
            l_p = constants[2]
            l_c = constants[3]
            field_off = constants[4]
            b_w = constants[5]

            # Precalculate offset.
            xi_off = l_p / 2 - l_c - field_off + (1 - b_w) * ct.c * t
            for i in prange(x.shape[0]):
                ez[i] = e_z_p * (xi[i] + xi_off)

        super().__init__(e_x=e_x, e_y=e_y, e_z=e_z)

    def _pre_gather(self, x, y, xi, t):
        n_p = self.density(t * ct.c, 0)
        w_p = ge.plasma_frequency(n_p * 1e-6)
        l_p = 2 * np.pi * ct.c / w_p
        g_x = w_p**2 / 2 * ct.m_e / (ct.e * ct.c)
        e_z_p = w_p**2 / 2 * ct.m_e / ct.e
        l_c = self.laser.xi_c
        b_w = self.laser.get_group_velocity(n_p)
        self.constants = np.array([g_x, e_z_p, l_p, l_c, self.field_offset, b_w])



from wake_t import GaussianPulse, PlasmaRamp, PlasmaStage, Beamline
from wake_t.utilities.bunch_generation import get_matched_bunch


def test_plasma_beamline():
    """
    This test checks that a beamline consisting of several consecutive plasma
    elements driven by the same laser works properly.

    At the moment, it simply checks that the simulation runs. More detailed
    checks should be added.

    """
    # Laser pulse
    laser = GaussianPulse(100e-6, l_0=800e-9, w_0=50e-6, a_0=3, tau=30e-15, z_foc=0)

    # Electron bunch
    bunch = get_matched_bunch(
        en_x=1e-6,
        en_y=1e-6,
        ene=200,
        ene_sp=0.1,
        s_t=1,
        xi_c=0,
        q_tot=10,
        n_part=1e4,
        n_p=1e23,
    )

    # Plasma elements
    wf_params = {
        "n_r": 200,
        "n_xi": 200,
        "dz_fields": 0.5e-3,
        "xi_min": -30e-6,
        "xi_max": 120e-6,
        "r_max": 200e-6,
        "laser": laser,
        "laser_evolution": True,
    }
    plateau = PlasmaStage(
        5e-3, 1e23, wakefield_model="quasistatic_2d", n_out=3, **wf_params
    )
    downramp = PlasmaRamp(
        5e-3,
        profile="inverse_square",
        ramp_type="downramp",
        wakefield_model="quasistatic_2d",
        plasma_dens_top=1e23,
        decay_length=1e-3,
        n_out=3,
        **wf_params,
    )
    plasma_beamline = Beamline([plateau, downramp])

    # Do tracking
    plasma_beamline.track(bunch)


if __name__ == "__main__":
    test_plasma_beamline()



import numpy as np

from wake_t.utilities.bunch_generation import get_gaussian_bunch_from_twiss
from wake_t import PlasmaRamp
from wake_t.diagnostics import analyze_bunch


def test_downramp():
    """
    Test that a plasma downramp changes the beta function of an initial beam
    as expected. The simplest `focusing_blowout` wakefield model is used in
    order to have a fast test.
    """
    # Set seed fo reproducible results.
    np.random.seed(0)

    # generate bunch
    en = 1e-6
    alpha = 0
    beta = 0.3e-3
    s_t = 3
    ene = 600
    ene_sp = 2
    bunch = get_gaussian_bunch_from_twiss(
        en_x=en,
        en_y=en,
        a_x=alpha,
        a_y=alpha,
        b_x=beta,
        b_y=beta,
        ene=ene,
        ene_sp=ene_sp,
        s_t=s_t,
        xi_c=0e-6,
        q_tot=100,
        n_part=1e4,
    )

    downramp = PlasmaRamp(
        1e-2,
        ramp_type="downramp",
        profile="inverse_square",
        plasma_dens_top=1e23,
        plasma_dens_down=1e21,
        wakefield_model="focusing_blowout",
        n_out=3,
    )

    downramp.track(bunch)
    bunch_params = analyze_bunch(bunch)
    beta_x = bunch_params["beta_x"]
    np.testing.assert_almost_equal(beta_x, 0.0097503087049851, decimal=9)


def test_upramp():
    """
    Test that a plasma upramp changes the beta function of an initial beam
    as expected. The simplest `focusing_blowout` wakefield model is used in
    order to have a fast test.
    """
    # Set seed fo reproducible results.
    np.random.seed(0)

    # generate bunch
    en = 1e-6
    alpha = 0
    beta = 1e-2
    s_t = 3
    ene = 600
    ene_sp = 2
    bunch = get_gaussian_bunch_from_twiss(
        en_x=en,
        en_y=en,
        a_x=alpha,
        a_y=alpha,
        b_x=beta,
        b_y=beta,
        ene=ene,
        ene_sp=ene_sp,
        s_t=s_t,
        xi_c=0e-6,
        q_tot=100,
        n_part=1e4,
    )

    downramp = PlasmaRamp(
        1e-2,
        ramp_type="upramp",
        profile="inverse_square",
        plasma_dens_top=1e23,
        plasma_dens_down=1e21,
        wakefield_model="focusing_blowout",
        n_out=3,
    )

    downramp.track(bunch)
    bunch_params = analyze_bunch(bunch)
    beta_x = bunch_params["beta_x"]
    np.testing.assert_almost_equal(beta_x, 0.0007631550504736755, decimal=9)


if __name__ == "__main__":
    test_downramp()
    test_upramp()



import numpy as np
import matplotlib.pyplot as plt
from pytest import approx

from wake_t import PlasmaStage
from wake_t.utilities.bunch_generation import get_matched_bunch
from wake_t.diagnostics import analyze_bunch_list


def test_multibunch_plasma_simulation(plot=False):
    """Test a plasma simulation with multiple electron bunches.

    This test checks that giving multiple bunches as an input to `track` works
    as expected, i.e., that all bunches are taken into account for the
    wakefield calculation and that all bunches are correctly evolved.
    """
    # Set numpy random seed to get reproducible results.
    np.random.seed(1)

    # Plasma density.
    n_p = 1e23

    # Create driver.
    en = 10e-6
    gamma = 2000
    s_t = 10
    ene_sp = 1
    q_tot = 500
    driver = get_matched_bunch(
        en_x=en,
        en_y=en,
        ene=gamma,
        ene_sp=ene_sp,
        s_t=s_t,
        xi_c=0,
        q_tot=q_tot,
        n_part=1e4,
        n_p=n_p,
    )

    # Create witness.
    en = 1e-6
    gamma = 200
    s_t = 3
    ene_sp = 0.1
    q_tot = 50
    witness = get_matched_bunch(
        en_x=en,
        en_y=en,
        ene=gamma,
        ene_sp=ene_sp,
        s_t=s_t,
        xi_c=-80e-6,
        q_tot=q_tot,
        n_part=1e4,
        n_p=n_p,
    )

    # Create plasma stage.
    plasma = PlasmaStage(
        length=1e-2,
        density=n_p,
        wakefield_model="quasistatic_2d",
        n_out=50,
        xi_max=20e-6,
        xi_min=-120e-6,
        r_max=70e-6,
        n_xi=280,
        n_r=70,
    )

    # Do tracking.
    output = plasma.track([driver, witness])

    # Analyze evolution.
    driver_params = analyze_bunch_list(output[0])
    witness_params = analyze_bunch_list(output[1])

    # Assert final parameters are correct.
    final_energy_driver = driver_params["avg_ene"][-1]
    final_energy_witness = witness_params["avg_ene"][-1]
    assert approx(final_energy_driver, rel=1e-10) == 1702.126922131719
    assert approx(final_energy_witness, rel=1e-9) == 636.7941326616957

    if plot:
        z = driver_params["prop_dist"] * 1e2
        plt.subplot(311)
        plt.plot(z, driver_params["emitt_x"] * 1e6, label="Driver", color="tab:blue")
        plt.plot(
            z, witness_params["emitt_x"] * 1e6, label="Witness", color="tab:orange"
        )
        plt.legend()
        plt.tick_params(axis="x", which="both", labelbottom=False)
        plt.ylabel("$\\epsilon_{nx}$ [$\\mu$m]")
        plt.subplot(312)
        plt.plot(z, driver_params["rel_ene_spread"] * 100, color="tab:blue")
        plt.plot(z, witness_params["rel_ene_spread"] * 100, color="tab:orange")
        plt.tick_params(axis="x", which="both", labelbottom=False)
        plt.ylabel("$\\frac{\\Delta \\gamma}{\\gamma}$ [%]")
        plt.subplot(313)
        plt.plot(z, driver_params["avg_ene"], color="tab:blue")
        plt.plot(z, witness_params["avg_ene"], color="tab:orange")
        plt.xlabel("z [cm]")
        plt.ylabel("$\\gamma$")
        plt.tight_layout()
        plt.show()


if __name__ == "__main__":
    test_multibunch_plasma_simulation(plot=True)



from copy import deepcopy

import numpy as np
import scipy.constants as ct

from wake_t import GaussianPulse, PlasmaStage, Beamline
from wake_t.utilities.bunch_generation import get_matched_bunch


def test_variable_parabolic_coefficient():
    """
    Checks that a z-dependent parabolic coefficient works as expected.
    A piecewise function for the coefficient is defined, which should be
    equivalent to tracking multiple plasma stages each having a different
    parabolic coefficient.
    """
    # Plasma density.
    n_p = 1e23

    # Grid and field parameters.
    xi_max = 0
    xi_min = -100e-6
    r_max = 100e-6
    Nxi = 100
    Nr = 100
    dz_fields = 1e-3

    # Laser parameters.
    a0 = 1
    w0 = 30e-6
    tau = 25e-15
    l0 = 0.8e-6

    # Guiding plasma channel
    r_e = ct.e**2 / (4.0 * np.pi * ct.epsilon_0 * ct.m_e * ct.c**2)
    rel_delta_n_over_w2 = 1.0 / (np.pi * r_e * w0**4 * n_p)

    # Length and parabolic coefficient of each section (stretch).
    L_stretches = [1e-2, 1e-2, 1e-2, 1e-2, 1e-2]
    pc_stretches = rel_delta_n_over_w2 * np.array([1, 2, 1, 0.5, 2])

    # Define z-dependent parabolic coefficient.
    def parabolic_coefficient(z):
        z_stretch_end = np.cumsum(L_stretches)
        i = np.sum(np.float32(z_stretch_end) <= np.float32(z))
        if i == len(L_stretches):
            i -= 1
        return pc_stretches[i]

    # Create identical laser pulses for each case.
    laser = GaussianPulse(-50e-6, a0, w0, tau, z_foc=0.0, l_0=l0)
    laser_1 = deepcopy(laser)
    laser_2 = deepcopy(laser)

    # Create identical bunches for each case.
    bunch = get_matched_bunch(
        1e-6, 1e-6, 200, 1, 3, laser.xi_c - 30e-6, 1e-6, 1e4, n_p=n_p
    )
    bunch_1 = bunch.copy()
    bunch_2 = bunch.copy()

    # Create single plasma stage (containing all sections).
    plasma_single = PlasmaStage(
        np.sum(L_stretches),
        n_p,
        wakefield_model="quasistatic_2d",
        n_out=10,
        xi_min=xi_min,
        xi_max=xi_max,
        r_max=r_max,
        r_max_plasma=r_max,
        laser=laser_1,
        laser_evolution=True,
        n_r=Nr,
        n_xi=Nxi,
        ppc=2,
        dz_fields=dz_fields,
        parabolic_coefficient=parabolic_coefficient,
    )

    # Track single plasma.
    plasma_single.track(bunch_1)

    # Create set of plasma stages, one per section.
    sub_stages = []
    for i, (length, pc) in enumerate(zip(L_stretches, pc_stretches)):
        stage = PlasmaStage(
            length,
            n_p,
            wakefield_model="quasistatic_2d",
            n_out=3,
            xi_min=xi_min,
            xi_max=xi_max,
            r_max=r_max,
            r_max_plasma=r_max,
            laser=laser_2,
            laser_evolution=True,
            n_r=Nr,
            n_xi=Nxi,
            ppc=2,
            dz_fields=dz_fields,
            parabolic_coefficient=pc,
        )
        sub_stages.append(stage)
    plasma_multi = Beamline(sub_stages)

    # Track through set of stages.
    plasma_multi.track(bunch_2)

    # Get final envelope of both lasers.
    a_env_1 = laser_1.get_envelope()
    a_mod_1 = np.abs(a_env_1)
    a_phase_1 = np.angle(a_env_1)
    a_env_2 = laser_2.get_envelope()
    a_mod_2 = np.abs(a_env_2)
    a_phase_2 = np.angle(a_env_2)

    # Check that both envelopes are equal.
    np.testing.assert_almost_equal(a_mod_2, a_mod_1)
    np.testing.assert_almost_equal(a_phase_2, a_phase_1, decimal=5)

    # Check final spot size value
    dr = r_max / Nr
    w0_final = calculate_spot_size(a_env_1, dr)
    np.testing.assert_almost_equal(w0_final, 2.03471741366461e-05)


def calculate_spot_size(a_env, dr):
    # Project envelope to r
    a_proj = np.sum(np.abs(a_env), axis=0)

    # Maximum is on axis
    a_max = a_proj[0]

    # Get first index of value below a_max / e
    i_first = np.where(a_proj <= a_max / np.e)[0][0]

    # Do linear interpolation to get more accurate value of w.
    # We build a line y = a + b*x, where:
    #     b = (y_2 - y_1) / (x_2 - x_1)
    #     a = y_1 - b*x_1
    #
    #     y_1 is the value of a_proj at i_first - 1
    #     y_2 is the value of a_proj at i_first
    #     x_1 and x_2 are the radial positions of y_1 and y_2
    #
    # We can then determine the spot size by interpolating between y_1 and y_2,
    # that is, do x = (y - a) / b, where y = a_max/e
    y_1 = a_proj[i_first - 1]
    y_2 = a_proj[i_first]
    x_1 = (i_first - 1) * dr + dr / 2
    x_2 = i_first * dr + dr / 2
    b = (y_2 - y_1) / (x_2 - x_1)
    a = y_1 - b * x_1
    w = (a_max / np.e - a) / b
    return w


if __name__ == "__main__":
    test_variable_parabolic_coefficient()



import numpy as np

from wake_t.physics_models.plasma_wakefields.qs_rz_baxevanis_ion.utils import (
    radial_gradient,
    longitudinal_gradient,
)


def test_gradients():
    """Check that the custom gradient methods behave as `np.gradient`."""
    nz = 200
    nr = 100

    z = np.linspace(0, 10, nz)
    r = np.linspace(0, 3, nr)
    dz = z[1] - z[0]
    dr = r[1] - r[0]
    z, r = np.meshgrid(z, r, indexing="ij")
    f = np.cos(z) * np.sin(r)

    dr_f = np.zeros((nz, nr))
    dz_f = np.zeros((nz, nr))

    longitudinal_gradient(f, dz, dz_f)
    radial_gradient(f, dr, dr_f)

    dz_f_np = np.gradient(f, dz, edge_order=2, axis=0)
    dr_f_np = np.gradient(
        np.concatenate((f[:, ::-1], f), axis=1), dr, edge_order=2, axis=1
    )[:, nr:]

    np.testing.assert_array_almost_equal(dz_f, dz_f_np)
    np.testing.assert_array_almost_equal(dr_f, dr_f_np)


if __name__ == "__main__":
    test_gradients()



"""
This module contais the tridiagonal matrix algorithm (TDMA) used by the laser
envelope solver.

Authors: Wilbert den Hertog, Ángel Ferran Pousa
"""

import numpy as np

from wake_t.utilities.numba import njit_serial


@njit_serial(fastmath=True)
def TDMA(a, b, c, d, p):
    """TriDiagonal Matrix Algorithm: solve a linear system Ax=b,
    where A is a tridiagonal matrix. Source:
    https://stackoverflow.com/questions/8733015/tridiagonal-matrix-algorithm-
    tdma-aka-thomas-algorithm-using-python-with-nump

    Parameters
    ----------
    a : array
        Lower diagonal of A. Dimension: nr-1.
    b : array
        Main diagonal of A. Dimension: nr.
    c : array
        Upper diagonal of A. Dimension: nr-1.
    d : array
        Solution vector. Dimension: nr.

    """
    n = len(d)
    w = np.empty(n - 1, dtype=np.complex128)
    g = np.empty(n, dtype=np.complex128)

    w[0] = c[0] / b[0]  # MAKE SURE THAT b[0]!=0
    g[0] = d[0] / b[0]

    for i in range(1, n - 1):
        a_im1 = a[i - 1]
        inv_coef = 1.0 / (b[i] - a_im1 * w[i - 1])
        g[i] = (d[i] - a_im1 * g[i - 1]) * inv_coef
        w[i] = c[i] * inv_coef

    g[-1] = (d[-1] - a[-2] * g[-2]) / (b[-1] - a[-2] * w[-2])

    # Fill in output array.
    p[-1] = g[-1]
    for i in range(n - 1, 0, -1):
        p[i - 1] = g[i - 1] - w[i - 1] * p[i]



"""Methods for gathering fields"""

from typing import List

import numpy as np

from wake_t.utilities.numba import njit_parallel, prange
from .base import Field


def gather_fields(
    fields: List[Field],
    x: np.ndarray,
    y: np.ndarray,
    z: np.ndarray,
    t: float,
    ex: np.ndarray,
    ey: np.ndarray,
    ez: np.ndarray,
    bx: np.ndarray,
    by: np.ndarray,
    bz: np.ndarray,
    bunch_name: str,
) -> None:
    """Gather all fields at the specified locations and time.

    Parameters
    ----------
    fields : list
        List of `Field`s.
    x : ndarray
        1D array containing the x position where to gather the fields.
    y : ndarray
        1D array containing the x position where to gather the fields.
    z : ndarray
        1D array containing the x position where to gather the fields.
    t : float
        Time at which the field is being gathered.
    ex : ndarray
        1D array where the gathered Ex values will be stored.
    ey : ndarray
        1D array where the gathered Ey values will be stored
    ez : ndarray
        1D array where the gathered Ez values will be stored
    bx : ndarray
        1D array where the gathered Bx values will be stored
    by : ndarray
        1D array where the gathered By values will be stored
    bz : ndarray
        1D array where the gathered Bz values will be stored
    bunch_name : str
        Name of the bunch that is gathering the fields
    """
    # Initially, set all field values to zero.
    reset_particle_fields(ex, ey, ez, bx, by, bz)

    # Gather contributions from all fields.
    for field in fields:
        field.gather(x, y, z, t, ex, ey, ez, bx, by, bz, bunch_name)


@njit_parallel
def reset_particle_fields(ex, ey, ez, bx, by, bz):
    """Set bunch field arrays to zero."""
    for i in prange(ex.size):
        ex[i] = 0.0
        ey[i] = 0.0
        ez[i] = 0.0
        bx[i] = 0.0
        by[i] = 0.0
        bz[i] = 0.0



import numpy as np
import scipy.constants as ct

from wake_t.fields.analytical_field import AnalyticalField
from wake_t.utilities.numba import prange


class CustomBlowoutWakefield(AnalyticalField):
    """
    Parameters
    ----------
    laser : LaserPulse
        Laser driver of the plasma stage.
    lon_field : float
        Value of the longitudinal electric field at the bunch center at the
        beginning of the tracking in units of V/m.
    lon_field_slope : float
        Value of the longitudinal electric field slope along z at the bunch
        center at the beginning of the tracking in units of V/m^2.
    foc_strength : float
        Value of the focusing gradient along the bunch in units of T/m.
    xi_fields : float
        Longitudinal position at which the wakefields have the values
        specified by the parameter above. By default, 0.

    """

    def __init__(
        self,
        n_p,
        laser,
        lon_field=None,
        lon_field_slope=None,
        foc_strength=None,
        xi_fields=0.0,
    ):
        super().__init__()
        self.density = n_p
        self.xi_fields = xi_fields
        self.laser = laser
        self.k = foc_strength
        self.e_z_0 = lon_field
        self.e_z_p = lon_field_slope

        def e_x(x, y, xi, t, ex, constants):
            k = constants[0]
            for i in prange(x.shape[0]):
                ex[i] = ct.c * k * x[i]

        def e_y(x, y, xi, t, ey, constants):
            k = constants[0]
            for i in prange(x.shape[0]):
                ey[i] = ct.c * k * y[i]

        def e_z(x, y, xi, t, ez, constants):
            e_z_0 = constants[1]
            e_z_p = constants[2]
            xi_fields = constants[3]
            b_w = constants[4]

            xi_off = -xi_fields + (1 - b_w) * ct.c * t
            for i in prange(x.shape[0]):
                ez[i] = e_z_0 + e_z_p * (xi[i] + xi_off)

        super().__init__(e_x=e_x, e_y=e_y, e_z=e_z)

    def _pre_gather(self, x, y, xi, t):
        n_p = self.density(t * ct.c, 0.0)
        b_w = self.laser.get_group_velocity(n_p)
        self.constants = np.array([self.k, self.e_z_0, self.e_z_p, self.xi_fields, b_w])



import math

from wake_t.utilities.numba import njit_serial


@njit_serial(error_model="numpy")
def gather_laser_sources(a2, nabla_a, r_min, r_max, dr, r, a2_pp, nabla_a_pp):
    """
    Gather the laser source terms (a2 and nabla_a) needed
    by the Baxevanis wakefield model into the plasma particles. This method is
    also catered to cylindrical geometry and assumes all plasma particles have
    the same longitudinal position (which is the case in a quasistatic model,
    where there is only a single column of particles).

    Parameters
    ----------
    a2, nabla_a : ndarray
        The source fields, corresponding respectively to a2 and nabla_a2.
        Each of them is a (nr+4) array, including 2 guard
        cells in each boundary.
    r_min, r_max : float
        Position of the first and last field values along r.
    dz : float
        Grid step size along the radial direction.
    r : 1darray
        Transverse position of the plasma particles.
    a2_pp, nabla_a_pp : ndarray
        Arrays where the gathered sources will be stored.

    """

    # Iterate over all particles.
    for i in range(r.shape[0]):
        # Get particle position.
        r_i = r[i]

        # Gather field only if particle is within field boundaries.
        if r_i <= r_max:
            # Position in cell units.
            r_i_cell = (r_i - r_min) / dr + 2

            # Indices of upper and lower cells in r and z.
            ir_lower = int(math.floor(r_i_cell))
            ir_upper = ir_lower + 1

            # If lower r cell is below axis, assume same value as first cell.
            # For `nabla_a2` and `b_theta_0`, invert the sign to ensure they
            # are `0` on axis.
            sign = 1
            if ir_lower < 2:
                ir_lower = 2
                sign = -1

            # Get field value at each bounding cell.
            fld_1_l = a2[ir_lower]
            fld_1_u = a2[ir_upper]
            fld_2_l = nabla_a[ir_lower] * sign
            fld_2_u = nabla_a[ir_upper]

            # Interpolate in r
            dr_u = ir_upper - r_i_cell
            dr_l = 1 - dr_u
            a2_pp[i] = dr_u * fld_1_l + dr_l * fld_1_u
            nabla_a_pp[i] = dr_u * fld_2_l + dr_l * fld_2_u
        else:
            a2_pp[i] = 0.0
            nabla_a_pp[i] = 0.0


@njit_serial(error_model="numpy")
def gather_bunch_sources(b_t, r_min, r_max, dr, r, b_t_pp):
    """
    Gathering the beam source terms (B_theta) needed by the Baxevanis
    wakefield model. This method is also
    catered to cylindrical geometry and assumes all plasma particles have the
    same longitudinal position (which is the case in a quasistatic model,
    where there is only a single column of particles).

    Parameters
    ----------
    b_t : ndarray
        The source term corresponding to the azimuthal magnetic field of a
        particle bunch. Array of size (nr+4) array, including 2 guard
        cells in each boundary.
    r_min, r_max : float
        Position of the first and last field values along r.
    dz : float
        Grid step size along the radial direction.
    r : 1darray
        Transverse position of the plasma particles.
    b_t_pp : ndarray
        Array where the gathered source will be stored.

    """

    # Iterate over all particles.
    for i in range(r.shape[0]):
        # Get particle position.
        r_i = r[i]

        # Position in cell units.
        r_i_cell = (r_i - r_min) / dr + 2

        # Indices of upper and lower cells in r and z.
        ir_lower = int(math.floor(r_i_cell))
        ir_upper = ir_lower + 1

        # If lower r cell is below axis, assume same value as first cell.
        # For `nabla_a2` and `b_theta_0`, invert the sign to ensure they
        # are `0` on axis.
        sign = 1
        if ir_lower < 2:
            ir_lower = 2
            sign = -1

        # Get field at lower and upper cell.
        # If the particle is within the boundaries of the grid, proceed
        # normally. Otherwise, calculate the lower and upper field
        # at "virtual" grid points. In principle this is not needed, because we
        # could avoid interpolation in this case and simply get the
        # field at the location of the particles using b_t[-1] * r_max / r_i.
        # This would be more exact. However, it makes it more difficult to
        # compare the results of simulations using an adaptive grid, which
        # would never be able to reproduce the exact same result as a case
        # with no adaptive grid due to the different (although more accurate)
        # field values.
        if r_i <= r_max:
            # Get field value at each bounding cell.
            fld_l = b_t[ir_lower] * sign
            fld_u = b_t[ir_upper]
        else:
            r_lower = (0.5 + ir_lower - 2) * dr
            r_upper = (0.5 + ir_upper - 2) * dr
            fld_l = b_t[-1] * r_max / r_lower * sign
            fld_u = b_t[-1] * r_max / r_upper

        # Interpolate in r
        dr_u = ir_upper - r_i_cell
        dr_l = 1 - dr_u
        b_t_pp[i] += dr_u * fld_l + dr_l * fld_u



import numpy as np
from wake_t.particles.deposition import deposit_3d_distribution


def test_uniform_plasma_deposition():
    """
    This test checks that a uniform distribution of plasma particles
    results in a uniform density distribution after being deposited into
    a grid.

    """
    # Grid and particle parameters.
    r_max = 10
    z_min = -5
    z_max = 5
    n_r = 20
    n_z = 21
    dr = r_max / n_r
    dz = (z_max - z_min) / (n_z - 1)
    r_fld = np.linspace(dr / 2, r_max - dr / 2, n_r)
    z_fld = np.linspace(z_min, z_max, n_z)
    parabolic_coefficient = 0.0
    ppc = 5
    r_max_plasma = r_max

    # Create plasma particles.
    dr_p = dr / ppc
    n_part = int(np.round(r_max_plasma / dr * ppc))
    r_max_plasma = n_part * dr_p
    r = np.linspace(dr_p / 2, r_max_plasma - dr_p / 2, n_part)
    pz = np.zeros_like(r)
    gamma = np.ones_like(r)
    q = dr_p * r + dr_p * parabolic_coefficient * r**3

    # Possible particle shapes.
    p_shapes = ["linear", "cubic"]

    # Test all shapes.
    for p_shape in p_shapes:
        # Allocate density array.
        rho_fld = np.zeros((n_z + 4, n_r + 4))

        # Deposit plasma column along the whole grid.
        for step in np.arange(n_z):
            i = -1 - step
            z_i = z_fld[i]
            w_rho = q / (dr * r * (1 - pz / gamma))
            z = np.full_like(r, z_i)
            x = r
            y = np.zeros_like(r)
            deposit_3d_distribution(
                z,
                x,
                y,
                w_rho,
                z_min,
                r_fld[0],
                n_z,
                n_r,
                dz,
                dr,
                rho_fld,
                p_shape=p_shape,
            )

        # Check array is uniform. Ignore last cells along r, as the edge of
        # plasma in the deposition array looks more smooth due to the
        # particle shape.
        if p_shape == "linear":
            assert np.sum(np.abs(rho_fld[2:-2, 2:-3] - 1.0)) < 1e-12
        elif p_shape == "cubic":
            assert np.sum(np.abs(rho_fld[2:-2, 2:-4] - 1.0)) < 1e-12


if __name__ == "__main__":
    test_uniform_plasma_deposition()



import math
import numpy as np
from aptools.plasma_accel.general_equations import (
    laser_radius_at_z_pos,
    laser_rayleigh_length,
)
from wake_t import GaussianPulse, FlattenedGaussianPulse


def test_gaussian_init():
    """Test initialization of Gaussian laser pulse.

    This test initializes a Gaussian laser pulse and checks that the spot size
    and a0 of the generated pulse agree with the desired values. This is done
    for a case on focus and a case off focus (z_foc = 1 cm).
    """
    # Grid properties.
    xi_max = 0e-6
    xi_min = -100e-6
    r_max = 100e-6
    nxi = 201
    nr = 200
    dr = r_max / nr

    # Time step (any value).
    dt = 100e-15

    # Laser properties.
    w0 = 30e-6
    a0 = 1
    l0 = 0.8e-6
    z_foc_to_test = [0.0, 1e-2]

    for z_foc in z_foc_to_test:
        # Create and initialize pulse.
        laser = GaussianPulse(
            xi_c=-50e-6, a_0=a0, w_0=w0, tau=25e-15, z_foc=z_foc, l_0=l0
        )
        laser.set_envelope_solver_params(xi_min, xi_max, r_max, nxi, nr, dt)
        laser.initialize_envelope()

        # Get complex envelope.
        a_env = laser.get_envelope()

        # Check correct w0.
        w_env = calculate_spot_size(a_env, dr)
        w_analytic = laser_radius_at_z_pos(w0, l0, z_foc)
        assert math.isclose(w_env, w_analytic, rel_tol=1e-4)

        # Check correct a0.
        z_r = laser_rayleigh_length(w0, l0)
        a0_analytic = a0 / np.sqrt(1 + (z_foc / z_r) ** 2)
        a0_env = np.max(np.abs(a_env))
        assert math.isclose(a0_env, a0_analytic, rel_tol=1e-4)


def test_flattened_gaussian_init():
    """Test initialization of flattened Gaussian laser pulse.

    This test initializes a flattend Gaussian laser pulse and checks that the
    spot size and a0 of the generated pulse agree with the desired values.
    The pulse is initialized with `N=0` so that it has a purely Gaussian shape,
    thus allowing for an easy comparison against theory. This is done
    for a case on focus and a case off focus (z_foc = 1 cm).
    """
    # Grid properties.
    xi_max = 0e-6
    xi_min = -100e-6
    r_max = 100e-6
    nxi = 201
    nr = 200
    dr = r_max / nr

    # Time step (any value).
    dt = 100e-15

    # Laser properties.
    w0 = 30e-6
    a0 = 1
    l0 = 0.8e-6
    z_foc_to_test = [0.0, 1e-2]

    for z_foc in z_foc_to_test:
        # Create and initialize pulse.
        laser = FlattenedGaussianPulse(
            xi_c=-50e-6, a_0=a0, w_0=w0, tau=25e-15, z_foc=z_foc, l_0=l0, N=0
        )
        laser.set_envelope_solver_params(xi_min, xi_max, r_max, nxi, nr, dt)
        laser.initialize_envelope()

        # Get complex envelope.
        a_env = laser.get_envelope()

        # Check correct w0.
        w_env = calculate_spot_size(a_env, dr)
        w_analytic = laser_radius_at_z_pos(w0, l0, -z_foc)
        assert math.isclose(w_env, w_analytic, rel_tol=1e-4)

        # Check correct a0.
        z_r = laser_rayleigh_length(w0, l0)
        a0_analytic = a0 / np.sqrt(1 + (z_foc / z_r) ** 2)
        a0_env = np.max(np.abs(a_env))
        assert math.isclose(a0_env, a0_analytic, rel_tol=1e-4)


def calculate_spot_size(a_env, dr):
    # Project envelope to r
    a_proj = np.sum(np.abs(a_env), axis=0)

    # Maximum is on axis
    a_max = a_proj[0]

    # Get first index of value below a_max / e
    i_first = np.where(a_proj <= a_max / np.e)[0][0]

    # Do linear interpolation to get more accurate value of w.
    # We build a line y = a + b*x, where:
    #     b = (y_2 - y_1) / (x_2 - x_1)
    #     a = y_1 - b*x_1
    #
    #     y_1 is the value of a_proj at i_first - 1
    #     y_2 is the value of a_proj at i_first
    #     x_1 and x_2 are the radial positions of y_1 and y_2
    #
    # We can then determine the spot size by interpolating between y_1 and y_2,
    # that is, do x = (y - a) / b, where y = a_max/e
    y_1 = a_proj[i_first - 1]
    y_2 = a_proj[i_first]
    x_1 = (i_first - 1) * dr + dr / 2
    x_2 = i_first * dr + dr / 2
    b = (y_2 - y_1) / (x_2 - x_1)
    a = y_1 - b * x_1
    w = (a_max / np.e - a) / b
    return w


if __name__ == "__main__":
    test_gaussian_init()
    test_flattened_gaussian_init()



"""Utilities for the laser envelope solver."""

import numpy as np
from wake_t.utilities.numba import njit_serial


@njit_serial
def unwrap(p, discont=None, axis=-1, period=6.283185307179586):
    """Numba version of numpy.unwrap.

    The implementation is taken from
    https://github.com/numba/numba/blob/main/numba/np/arraymath.py,
    which currently is not yet included in the latest Numba release.
    """
    if axis != -1:
        msg = 'Value for argument "axis" is not supported'
        raise ValueError(msg)
    # Flatten to a 2D array, keeping axis -1
    p_init = np.asarray(p).astype(np.float64)
    init_shape = p_init.shape
    last_axis = init_shape[-1]
    p_new = p_init.reshape((p_init.size // last_axis, last_axis))
    # Manipulate discont and period
    if discont is None:
        discont = period / 2
    interval_high = period / 2
    boundary_ambiguous = True
    interval_low = -interval_high

    slice1 = (slice(1, None, None),)

    # Work on each row separately
    for i in range(p_init.size // last_axis):
        row = p_new[i]
        dd = np.diff(row)
        ddmod = np.mod(dd - interval_low, period) + interval_low
        if boundary_ambiguous:
            ddmod = np.where((ddmod == interval_low) & (dd > 0), interval_high, ddmod)
        ph_correct = ddmod - dd

        ph_correct = np.where(np.array([abs(x) for x in dd]) < discont, 0, ph_correct)
        ph_ravel = np.where(np.array([abs(x) for x in dd]) < discont, 0, ph_correct)
        ph_correct = np.reshape(ph_ravel, ph_correct.shape)
        up = np.copy(row)
        up[slice1] = row[slice1] + ph_correct.cumsum()
        p_new[i] = up

    return p_new.reshape(init_shape)



from wake_t import PlasmaStage, GaussianPulse


def test_no_bunch_plasma_simulation():
    """Test a plasma simulation with no electron bunches (only a laser).

    This test checks that a plasma simulation can run without giving
    any bunch as input to `track`.
    """
    # Plasma density.
    n_p = 1e23

    # Create driver.
    laser = GaussianPulse(0.0, l_0=800e-9, w_0=30e-6, a_0=3, tau=30e-15, z_foc=0.0)

    # Create plasma stage.
    plasma = PlasmaStage(
        length=1e-2,
        density=n_p,
        wakefield_model="quasistatic_2d",
        n_out=50,
        xi_max=20e-6,
        xi_min=-120e-6,
        r_max=200e-6,
        n_xi=280,
        n_r=200,
        laser=laser,
    )

    # Do simulation with no electron bunch.
    plasma.track()


if __name__ == "__main__":
    test_no_bunch_plasma_simulation()



__version__ = "0.9.1"


from .beamline_elements import (
    PlasmaStage,
    PlasmaRamp,
    ActivePlasmaLens,
    Drift,
    Dipole,
    Quadrupole,
    Sextupole,
    Beamline,
)
from .physics_models.collective_effects.csr import set_csr_settings
from .physics_models.laser.laser_pulse import (
    GaussianPulse,
    LaguerreGaussPulse,
    FlattenedGaussianPulse,
)
from .particles.particle_bunch import ParticleBunch


__all__ = [
    "__version__",
    "PlasmaStage",
    "PlasmaRamp",
    "ActivePlasmaLens",
    "Drift",
    "Dipole",
    "Quadrupole",
    "Sextupole",
    "Beamline",
    "set_csr_settings",
    "GaussianPulse",
    "LaguerreGaussPulse",
    "FlattenedGaussianPulse",
    "ParticleBunch",
]



import os
from copy import deepcopy

import numpy as np
import matplotlib.pyplot as plt

from wake_t import PlasmaStage
from wake_t.utilities.bunch_generation import get_matched_bunch
from wake_t.diagnostics import analyze_bunch_list
from wake_t.physics_models.plasma_wakefields.qs_rz_baxevanis_ion.gather import (
    gather_bunch_sources,
)
from wake_t.fields.gather import gather_fields


TEST_DIR = os.path.dirname(os.path.abspath(__file__))


def test_adaptive_grid():
    """Test a plasma simulation using adaptive grids.

    This test runs a simulation of a single time step with and
    without using adaptive grids. The adaptive grids have the same
    radial extent and resolution as the base grid. As such, the test
    checks that the simulation results are identical.

    The resolution of the adaptive and the base grids is identical.
    """
    # Set numpy random seed to get reproducible results.
    np.random.seed(1)

    # Plasma density.
    n_p = 1e23

    # Create driver.
    en = 10e-6
    gamma = 2000
    s_t = 10
    ene_sp = 1
    q_tot = 500
    driver = get_matched_bunch(
        en_x=en,
        en_y=en,
        ene=gamma,
        ene_sp=ene_sp,
        s_t=s_t,
        xi_c=0,
        q_tot=q_tot,
        n_part=1e4,
        n_p=n_p,
        name="driver",
    )

    # Create witness.
    en = 1e-6
    gamma = 200
    s_t = 3
    ene_sp = 0.1
    q_tot = 50
    witness = get_matched_bunch(
        en_x=en,
        en_y=en,
        ene=gamma,
        ene_sp=ene_sp,
        s_t=s_t,
        xi_c=-80e-6,
        q_tot=q_tot,
        n_part=1e4,
        n_p=n_p,
        name="witness",
    )

    # Run simulations with and without adaptive grid.
    driver_params, witness_params, plasma = run_simulation(
        deepcopy(driver), deepcopy(witness), n_p, use_ag=False
    )
    driver_params_ag, witness_params_ag, plasma_ag = run_simulation(
        deepcopy(driver), deepcopy(witness), n_p, use_ag=True
    )

    ag_driver = plasma_ag.wakefield.bunch_grids["driver"]
    ag_witness = plasma_ag.wakefield.bunch_grids["witness"]
    ags = [ag_driver, ag_witness]
    q_bunch_base = plasma.wakefield.q_bunch[2:-2, 2:-2]
    bt_bunch_base = plasma.wakefield.b_t_bunch[2:-2, 2:-2]
    er_base = plasma.wakefield.fld_arrays[4][2:-2, 2:-2]
    ez_base = plasma.wakefield.fld_arrays[5][2:-2, 2:-2]
    bt_base = plasma.wakefield.fld_arrays[6][2:-2, 2:-2]

    for ag in ags:
        q_bunch_ag = ag.q_bunch[2:-2, 2:-2]
        bt_bunch_ag = ag.b_t_bunch[2:-2, 2:-2]
        er_ag = ag.e_r[2:-2, 2:-2]
        ez_ag = ag.e_z[2:-2, 2:-2]
        bt_ag = ag.b_t[2:-2, 2:-2]

        # Check that the shape of the grid fields and properties is consistent.
        assert er_ag.shape[0] == ag.xi_grid.shape[0]
        assert er_ag.shape[1] == ag.r_grid.shape[0]
        assert er_ag.shape[1] == ag.nr

        # Check that the grid spacing is the same as in the base grid.
        assert plasma_ag.wakefield.dr == ag.dr
        assert plasma_ag.wakefield.dxi == ag.dxi

        # Check that the grid coordinate arrays overlap.
        np.testing.assert_array_equal(
            plasma_ag.wakefield.r_fld, ag.r_grid[: -ag.nr_border]
        )
        np.testing.assert_array_equal(plasma_ag.wakefield.xi_fld[ag.i_grid], ag.xi_grid)

        # Check that bunch charge distribution and space charge agree
        # between the base and the adaptive grid.
        np.testing.assert_allclose(
            q_bunch_base[ag.i_grid], q_bunch_ag[:, : -ag.nr_border], rtol=1e-10
        )
        np.testing.assert_allclose(
            bt_bunch_base[ag.i_grid], bt_bunch_ag[:, : -ag.nr_border], rtol=1e-11
        )

        # Check that the field in the adaptive grid agree with those
        # of the base grid.
        np.testing.assert_allclose(
            bt_base[ag.i_grid], bt_ag[:, : -ag.nr_border], rtol=1e-8
        )
        np.testing.assert_allclose(
            er_base[ag.i_grid, :-1], er_ag[:, : -ag.nr_border - 1], rtol=1e-8
        )
        np.testing.assert_allclose(
            ez_base[ag.i_grid][1:-1], ez_ag[1:-1, : -ag.nr_border], rtol=1e-8
        )


def test_adaptive_grid_undersized():
    """Test a plasma simulation using adaptive grids.

    This test runs a simulation of a single time step with and
    without using adaptive grids. The adaptive grids are undersized,
    meaning that the bunches will not fully fit within them. As such, the
    beams have to deposit to and gather from both the adaptive grids
    as well as the base grid. This test checks that the simulation results
    when the adaptive grids are undersized are identical to when no
    adaptive grids are used.

    The resolution of the adaptive and the base grids is identical.
    """
    # Set numpy random seed to get reproducible results.
    np.random.seed(1)

    # Plasma density.
    n_p = 1e23

    # Create driver.
    en = 10e-6
    gamma = 2000
    s_t = 10
    ene_sp = 1
    q_tot = 10
    driver = get_matched_bunch(
        en_x=en,
        en_y=en,
        ene=gamma,
        ene_sp=ene_sp,
        s_t=s_t,
        xi_c=0,
        q_tot=q_tot,
        n_part=1e4,
        n_p=n_p,
        name="driver",
    )

    # Create witness.
    en = 1e-6
    gamma = 200
    s_t = 3
    ene_sp = 0.1
    q_tot = 5
    witness = get_matched_bunch(
        en_x=en,
        en_y=en,
        ene=gamma,
        ene_sp=ene_sp,
        s_t=s_t,
        xi_c=-60e-6,
        q_tot=q_tot,
        n_part=1e4,
        n_p=n_p,
        name="witness",
    )

    for p_shape in ["linear", "cubic"]:
        # Run simulations with and without adaptive grid.
        driver_params, witness_params, plasma = run_simulation(
            deepcopy(driver),
            deepcopy(witness),
            n_p,
            length=1e-6,
            use_ag=False,
            p_shape=p_shape,
        )

        # Run simulations with fixed (undersized) adaptive grid.
        driver_params_ag, witness_params_ag, plasma_ag = run_simulation(
            deepcopy(driver),
            deepcopy(witness),
            n_p,
            length=1e-6,
            use_ag=True,
            nr_ag=[3, 3],
            r_max_ag=[3e-6, 3e-6],
            r_lim_ag=[None, None],
            p_shape=p_shape,
        )
        # plt.show()
        ag_driver = plasma_ag.wakefield.bunch_grids["driver"]
        ag_witness = plasma_ag.wakefield.bunch_grids["witness"]
        q_bunch_base_ag = plasma_ag.wakefield.q_bunch[2:-2, 2:-2]
        q_bunch_base = plasma.wakefield.q_bunch[2:-2, 2:-2]
        er_base_ag = plasma_ag.wakefield.e_r[2:-2, 2:-2]
        ez_base_ag = plasma_ag.wakefield.e_z[2:-2, 2:-2]
        bt_base_ag = plasma_ag.wakefield.b_t[2:-2, 2:-2]
        er_base = plasma.wakefield.e_r[2:-2, 2:-2]
        ez_base = plasma.wakefield.e_z[2:-2, 2:-2]
        bt_base = plasma.wakefield.b_t[2:-2, 2:-2]

        q_bunch_ag_combined = deepcopy(q_bunch_base_ag)
        ags = [ag_driver, ag_witness]
        for ag in ags:
            q_bunch_ag = ag.q_bunch[2:-2, 2:]

            q_bunch_ag_combined[ag.i_grid, : ag.nr + 2] += q_bunch_ag

        # Check that the charge has been correctly deposited across all grids.
        np.testing.assert_allclose(q_bunch_base, q_bunch_ag_combined, rtol=1e-12)

        r_test = np.linspace(0, 70e-6, 70)
        b_t_test = np.zeros_like(r_test)
        b_t_test_ag = np.zeros_like(r_test)
        slice_i = 180
        gather_bunch_sources(
            plasma.wakefield.b_t_bunch[slice_i + 2],
            plasma.wakefield.r_fld[0],
            plasma.wakefield.r_fld[-1] + 2 * plasma.wakefield.dr,
            plasma.wakefield.dr,
            r_test,
            b_t_test,
        )

        gather_bunch_sources(
            ag_driver.b_t_bunch[slice_i + 2 - ag_driver.i_grid[0]],
            ag_driver.r_grid[0],
            ag_driver.r_grid[-1] + 2 * ag_driver.dr,
            ag_driver.dr,
            r_test,
            b_t_test_ag,
        )
        gather_bunch_sources(
            plasma_ag.wakefield.b_t_bunch[slice_i + 2],
            plasma_ag.wakefield.r_fld[0],
            plasma_ag.wakefield.r_fld[-1] + 2 * plasma_ag.wakefield.dr,
            plasma_ag.wakefield.dr,
            r_test,
            b_t_test_ag,
        )
        np.testing.assert_allclose(b_t_test, b_t_test_ag, rtol=1e-12)

        # Check that the fields in the base grid agree between the cases with
        # and without adaptive grids.
        np.testing.assert_allclose(er_base, er_base_ag, rtol=1e-7)
        np.testing.assert_allclose(ez_base, ez_base_ag, rtol=1e-5)
        np.testing.assert_allclose(bt_base, bt_base_ag, rtol=1e-7)

        # Check that the fields gathered by the bunch agree between the cases
        # with and without adaptive grids.
        field_arrays = deepcopy(driver).get_field_arrays()
        gather_fields(
            [plasma.wakefield],
            driver.x,
            driver.y,
            driver.xi,
            0.0,
            *field_arrays,
            driver.name,
        )

        field_arrays_ag = deepcopy(driver).get_field_arrays()
        gather_fields(
            [plasma_ag.wakefield],
            driver.x,
            driver.y,
            driver.xi,
            0.0,
            *field_arrays_ag,
            driver.name,
        )
        for arr, arr_ag in zip(field_arrays, field_arrays_ag):
            np.testing.assert_allclose(arr, arr_ag, rtol=1e-10)


def test_adaptive_grids_evolution(create_test_data=False, plot=False):
    """Test that the radial evolution of the adaptive grids is as expected."""
    # Set numpy random seed to get reproducible results.
    np.random.seed(1)

    # Plasma density.
    n_p = 1e23

    # Create driver.
    en = 10e-6
    gamma = 2000
    s_t = 10
    ene_sp = 1
    q_tot = 500
    driver = get_matched_bunch(
        en_x=en,
        en_y=en,
        ene=gamma,
        ene_sp=ene_sp,
        s_t=s_t,
        xi_c=0,
        q_tot=q_tot,
        n_part=1e4,
        n_p=n_p,
        name="driver",
    )

    # Create witness.
    en = 1e-6
    gamma = 200
    s_t = 3
    ene_sp = 0.1
    q_tot = 50
    witness = get_matched_bunch(
        en_x=en,
        en_y=en,
        ene=gamma,
        ene_sp=ene_sp,
        s_t=s_t,
        xi_c=-80e-6,
        q_tot=q_tot,
        n_part=1e4,
        n_p=n_p,
        name="witness",
    )

    # Run simulation without adaptive grid.
    driver_params, witness_params, plasma = run_simulation(
        deepcopy(driver), deepcopy(witness), n_p, use_ag=False, length=1e-2
    )
    # Run simulation with adaptive grid.
    driver_params_ag, witness_params_ag, plasma_ag = run_simulation(
        deepcopy(driver),
        deepcopy(witness),
        n_p,
        length=1e-2,
        use_ag=True,
        nr_ag=[6, 6],
        r_max_ag=[None, None],
        r_lim_ag=[None, None],
    )
    # Run simulation with fixed adaptive grid.
    driver_params_agf, witness_params_agf, plasma_agf = run_simulation(
        deepcopy(driver),
        deepcopy(witness),
        n_p,
        length=1e-2,
        use_ag=True,
        nr_ag=[6, 6],
        r_max_ag=[6e-6, 6e-6],
        r_lim_ag=[None, None],
    )
    # Run simulation with limited adaptive grid.
    driver_params_agl, witness_params_agl, plasma_agl = run_simulation(
        deepcopy(driver),
        deepcopy(witness),
        n_p,
        length=1e-2,
        use_ag=True,
        nr_ag=[6, 6],
        r_max_ag=[None, None],
        r_lim_ag=[15e-6, 15e-6],
    )

    # Check that the fixed grid is indeed fixed
    np.testing.assert_array_equal(
        plasma_agf.wakefield.bunch_grids["driver"]._r_max_hist, 6e-6
    )
    np.testing.assert_array_equal(
        plasma_agf.wakefield.bunch_grids["witness"]._r_max_hist, 6e-6
    )

    # Save arrays with the evolution of the radial size of the grids.
    if create_test_data:
        np.savetxt(
            os.path.join(TEST_DIR, "resources", "r_max_hist_driver_ag.txt"),
            plasma_ag.wakefield.bunch_grids["driver"]._r_max_hist,
        )
        np.savetxt(
            os.path.join(TEST_DIR, "resources", "r_max_hist_witness_ag.txt"),
            plasma_ag.wakefield.bunch_grids["witness"]._r_max_hist,
        )
        np.savetxt(
            os.path.join(TEST_DIR, "resources", "r_max_hist_driver_agl.txt"),
            plasma_agl.wakefield.bunch_grids["driver"]._r_max_hist,
        )
        np.savetxt(
            os.path.join(TEST_DIR, "resources", "r_max_hist_witness_agl.txt"),
            plasma_agl.wakefield.bunch_grids["witness"]._r_max_hist,
        )

    # Check that the radial evolution is as expected.
    np.testing.assert_allclose(
        plasma_ag.wakefield.bunch_grids["driver"]._r_max_hist,
        np.loadtxt(os.path.join(TEST_DIR, "resources", "r_max_hist_driver_ag.txt")),
        rtol=1e-12,
    )
    np.testing.assert_allclose(
        plasma_ag.wakefield.bunch_grids["witness"]._r_max_hist,
        np.loadtxt(os.path.join(TEST_DIR, "resources", "r_max_hist_witness_ag.txt")),
        rtol=1e-12,
    )
    np.testing.assert_allclose(
        plasma_agl.wakefield.bunch_grids["driver"]._r_max_hist,
        np.loadtxt(os.path.join(TEST_DIR, "resources", "r_max_hist_driver_agl.txt")),
        rtol=1e-12,
    )
    np.testing.assert_allclose(
        plasma_agl.wakefield.bunch_grids["witness"]._r_max_hist,
        np.loadtxt(os.path.join(TEST_DIR, "resources", "r_max_hist_witness_agl.txt")),
        rtol=1e-12,
    )

    # Check that the beam evolution is identical for a fixed adaptive grid
    # and no adaptive grid.
    for param in driver_params.keys():
        np.testing.assert_allclose(
            driver_params[param], driver_params_agf[param], rtol=1e-12
        )
        np.testing.assert_allclose(
            witness_params[param], witness_params_agf[param], rtol=1e-11
        )

    if plot:
        plt.plot(
            plasma_ag.wakefield.bunch_grids["driver"]._r_max_hist, label="driver grid"
        )
        plt.plot(
            plasma_ag.wakefield.bunch_grids["witness"]._r_max_hist, label="witness grid"
        )
        plt.plot(
            plasma_agl.wakefield.bunch_grids["driver"]._r_max_hist,
            label="driver grid (limited)",
        )
        plt.plot(
            plasma_agl.wakefield.bunch_grids["witness"]._r_max_hist,
            label="witness grid (limited)",
        )
        plt.legend()
        plt.show()


def run_simulation(
    driver,
    witness,
    n_p,
    length=1e-6,
    use_ag=False,
    nr_ag=[70, 70],
    r_max_ag=[70e-6, 70e-6],
    r_lim_ag=[None, None],
    p_shape="cubic",
):
    model_params = {
        "xi_max": 20e-6,
        "xi_min": -90e-6,
        "r_max": 70e-6,
        "n_xi": 220,
        "n_r": 70,
        "dz_fields": 1e-3,
        "p_shape": p_shape,
    }
    if use_ag:
        model_params["use_adaptive_grids"] = True
        model_params["adaptive_grid_nr"] = nr_ag
        model_params["adaptive_grid_r_max"] = r_max_ag
        model_params["adaptive_grid_r_lim"] = r_lim_ag

    plasma = PlasmaStage(
        length=length,
        density=n_p,
        wakefield_model="quasistatic_2d_ion",
        n_out=3,
        **model_params,
    )

    # Do tracking.
    output = plasma.track([driver, witness])

    # Analyze evolution.
    driver_params = analyze_bunch_list(output[0])
    witness_params = analyze_bunch_list(output[1])
    return driver_params, witness_params, plasma


if __name__ == "__main__":
    test_adaptive_grid()
    test_adaptive_grid_undersized()
    test_adaptive_grids_evolution(create_test_data=True, plot=True)



"""Contains the class used to define analytic fields."""

from typing import Callable, Optional, List
import numpy as np

from .base import Field
from wake_t.utilities.numba import njit_parallel


# Define type alias.
FieldFunction = Callable[
    [np.ndarray, np.ndarray, np.ndarray, float, np.ndarray, List], np.ndarray
]


class AnalyticalField(Field):
    """Class used to define fields with analytical components.

    The given components (Ex, Ey, Ez, Bx, By, Bz) must be functions taking 5
    arguments (3 arrays containing the x, y, z positions where to calculate the
    field; 1 array with the same size where the field values will be stored;
    and a list of constants). The given functions must be written in a way
    which allows them to be compiled with ``numba``.

    Not all components need to be given. Those which are not specified will
    simply return a zero array when gathered.

    In addition to the field components, a list of constants can also be given.
    This list of constants is always passed to the field functions and can be
    used to compute the field.

    Parameters
    ----------
    e_x : callable, optional
        Function defining the Ex component.
    e_y : callable, optional
        Function defining the Ey component.
    e_z : callable, optional
        Function defining the Ez component.
    b_x : callable, optional
        Function defining the Bx component.
    b_y : callable, optional
        Function defining the By component.
    b_z : callable, optional
        Function defining the Bz component.
    constants : list, optional
        List of constants to be passed to each component.

    Examples
    --------
    >>> from numba import prange
    >>> def linear_ex(x, y, z, t, ex, constants):
    ...     ex_slope = constants[0]
    ...     for i in prange(x.shape[0]):
    ...         ex[i] = ex_slope * x[i]
    ...
    >>> ex = AnalyticField(e_x=linear_ex, constants=[1e6])

    """

    def __init__(
        self,
        e_x: Optional[FieldFunction] = None,
        e_y: Optional[FieldFunction] = None,
        e_z: Optional[FieldFunction] = None,
        b_x: Optional[FieldFunction] = None,
        b_y: Optional[FieldFunction] = None,
        b_z: Optional[FieldFunction] = None,
        constants: Optional[List] = None,
    ) -> None:
        super().__init__()

        constants = [] if constants is None else constants

        def no_field(x, y, z, t, fld, k):
            """Default field component."""
            pass

        self.__e_x = njit_parallel(e_x) if e_x is not None else no_field
        self.__e_y = njit_parallel(e_y) if e_y is not None else no_field
        self.__e_z = njit_parallel(e_z) if e_z is not None else no_field
        self.__b_x = njit_parallel(b_x) if b_x is not None else no_field
        self.__b_y = njit_parallel(b_y) if b_y is not None else no_field
        self.__b_z = njit_parallel(b_z) if b_z is not None else no_field
        self.constants = np.array(constants)

    def _pre_gather(self, x, y, z, t):
        """Function that is automatically called just before gathering.

        This method can be overwritten by derived classes and used to,
        for example, pre-compute any useful quantities. This method is not
        compiled by numba.
        """
        pass

    def _gather(self, x, y, z, t, ex, ey, ez, bx, by, bz, bunch_name):
        self._pre_gather(x, y, z, t)
        self.__e_x(x, y, z, t, ex, self.constants)
        self.__e_y(x, y, z, t, ey, self.constants)
        self.__e_z(x, y, z, t, ez, self.constants)
        self.__b_x(x, y, z, t, bx, self.constants)
        self.__b_y(x, y, z, t, by, self.constants)
        self.__b_z(x, y, z, t, bz, self.constants)



import numpy as np
import scipy.constants as ct

from wake_t.beamline_elements import FieldQuadrupole, Quadrupole
from wake_t.utilities.bunch_generation import get_gaussian_bunch_from_twiss


def test_field_vs_tm_quadrupole():
    """
    This test checks that the FieldElement-based quadrupole (FieldQuadrupole)
    and the TM quadrupole (Quadrupole) produce similar results.
    """

    emitt_nx = emitt_ny = 1e-6  # m
    beta_x = beta_y = 1.0  # m
    s_t = 100.0  # fs
    gamma_avg = 1000
    ene_spread = 0.1  # %
    q_bunch = 30  # pC
    xi_avg = 0.0  # m
    n_part = 1e4
    bunch_1 = get_gaussian_bunch_from_twiss(
        en_x=emitt_nx,
        en_y=emitt_ny,
        a_x=0,
        a_y=0,
        b_x=beta_x,
        b_y=beta_y,
        ene=gamma_avg,
        ene_sp=ene_spread,
        s_t=s_t,
        xi_c=xi_avg,
        q_tot=q_bunch,
        n_part=n_part,
        name="elec_bunch",
    )

    bunch_2 = bunch_1.copy()

    foc_strength = 100  # T/m
    quadrupole_length = 0.05  # m
    k1 = foc_strength * ct.e / ct.m_e / ct.c / gamma_avg

    field_quadrupole = FieldQuadrupole(quadrupole_length, foc_strength)
    tm_quadrupole = Quadrupole(quadrupole_length, k1)

    field_quadrupole.track(bunch_1)
    tm_quadrupole.track(bunch_2)

    np.testing.assert_allclose(bunch_1.x, bunch_2.x, rtol=1e-3, atol=1e-7)
    np.testing.assert_allclose(bunch_1.y, bunch_2.y, rtol=1e-3, atol=1e-7)
    np.testing.assert_allclose(bunch_1.xi, bunch_2.xi, rtol=1e-3, atol=1e-7)
    np.testing.assert_allclose(bunch_1.px, bunch_2.px, rtol=1e-3, atol=1e-3)
    np.testing.assert_allclose(bunch_1.py, bunch_2.py, rtol=1e-3, atol=1e-3)
    np.testing.assert_allclose(bunch_1.pz, bunch_2.pz, rtol=1e-3, atol=1e-3)


if __name__ == "__main__":
    test_field_vs_tm_quadrupole()



"""Contains the definition of the `PlasmaParticles` class."""

import numpy as np


class PlasmaParticles:
    """
    Class containing the 1D slice of plasma particles used in the quasi-static
    Baxevanis wakefield model.

    Parameters
    ----------
    r_max : float
        Maximum radial extension of the simulation box in normalized units.
    r_max_plasma : float
        Maximum radial extension of the plasma column in normalized units.
    parabolic_coefficient : float
        The coefficient for the transverse parabolic density profile.
    dr : float
        Radial step size of the discretized simulation box.
    ppc : int
        Number of particles per cell.
    pusher : str
        Particle pusher used to evolve the plasma particles. Possible
        values are `'rk4'` and `'ab5'`.

    """

    def __init__(self, r_max, r_max_plasma, parabolic_coefficient, dr, ppc, pusher):
        # Calculate total number of plasma particles.
        n_part = int(np.round(r_max_plasma / dr * ppc))

        # Readjust plasma extent to match number of particles.
        dr_p = dr / ppc
        r_max_plasma = n_part * dr_p

        # Store parameters.
        self.r_max = r_max
        self.r_max_plasma = r_max_plasma
        self.parabolic_coefficient = parabolic_coefficient
        self.dr = dr
        self.ppc = ppc
        self.dr_p = dr / ppc
        self.pusher = pusher
        self.n_part = n_part

    def initialize(self):
        """Initialize column of plasma particles."""

        # Initialize particle arrays.
        self.r = np.linspace(
            self.dr_p / 2, self.r_max_plasma - self.dr_p / 2, self.n_part
        )
        self.pr = np.zeros(self.n_part)
        self.pz = np.zeros(self.n_part)
        self.gamma = np.ones(self.n_part)
        self.q = self.dr_p * self.r + self.dr_p * self.parabolic_coefficient * self.r**3

        # Allocate arrays that will contain the fields experienced by the
        # particles.
        self.allocate_field_arrays()

        # Allocate arrays needed for the particle pusher.
        if self.pusher == "ab5":
            self.allocate_ab5_arrays()
        elif self.pusher == "rk4":
            self.allocate_rk4_arrays()
            self.allocate_rk4_field_arrays()

    def allocate_field_arrays(self):
        """Allocate arrays for the fields experienced by the particles.

        In order to evolve the particles to the next longitudinal position,
        it is necessary to know the fields that they are experiencing. These
        arrays are used for storing the value of these fields at the location
        of each particle.
        """
        self.__a2 = np.zeros(self.n_part)
        self.__nabla_a2 = np.zeros(self.n_part)
        self.__b_t_0 = np.zeros(self.n_part)
        self.__b_t = np.zeros(self.n_part)
        self.__psi = np.zeros(self.n_part)
        self.__dr_psi = np.zeros(self.n_part)
        self.__dxi_psi = np.zeros(self.n_part)
        self.__field_arrays = [
            self.__a2,
            self.__nabla_a2,
            self.__b_t_0,
            self.__b_t,
            self.__psi,
            self.__dr_psi,
            self.__dxi_psi,
        ]

    def get_field_arrays(self):
        """Get arrays containing the fields experienced by the particles."""
        return self.__field_arrays

    def allocate_ab5_arrays(self):
        """Allocate the arrays needed for the 5th order Adams-Bashforth pusher.

        The AB5 pusher needs the derivatives of r and pr for each particle
        at the last 5 plasma slices. This method allocates the arrays that will
        store these derivatives.
        """
        self.__dr_1 = np.zeros(self.n_part)
        self.__dr_2 = np.zeros(self.n_part)
        self.__dr_3 = np.zeros(self.n_part)
        self.__dr_4 = np.zeros(self.n_part)
        self.__dr_5 = np.zeros(self.n_part)
        self.__dpr_1 = np.zeros(self.n_part)
        self.__dpr_2 = np.zeros(self.n_part)
        self.__dpr_3 = np.zeros(self.n_part)
        self.__dpr_4 = np.zeros(self.n_part)
        self.__dpr_5 = np.zeros(self.n_part)
        self.__dr_arrays = [
            self.__dr_1,
            self.__dr_2,
            self.__dr_3,
            self.__dr_4,
            self.__dr_5,
        ]
        self.__dpr_arrays = [
            self.__dpr_1,
            self.__dpr_2,
            self.__dpr_3,
            self.__dpr_4,
            self.__dpr_5,
        ]

    def get_ab5_arrays(self):
        """Get the arrays needed by the 5th order Adams-Bashforth pusher."""
        return self.__dr_arrays, self.__dpr_arrays

    def allocate_rk4_arrays(self):
        """Allocate the arrays needed for the 4th order Runge-Kutta pusher.

        The RK4 pusher needs the derivatives of r and pr for each particle at
        the current slice and at 3 intermediate substeps. This method allocates
        the arrays that will store these derivatives.
        """
        self.__dr_1 = np.zeros(self.n_part)
        self.__dr_2 = np.zeros(self.n_part)
        self.__dr_3 = np.zeros(self.n_part)
        self.__dr_4 = np.zeros(self.n_part)
        self.__dpr_1 = np.zeros(self.n_part)
        self.__dpr_2 = np.zeros(self.n_part)
        self.__dpr_3 = np.zeros(self.n_part)
        self.__dpr_4 = np.zeros(self.n_part)
        self.__dr_arrays = [self.__dr_1, self.__dr_2, self.__dr_3, self.__dr_4]
        self.__dpr_arrays = [self.__dpr_1, self.__dpr_2, self.__dpr_3, self.__dpr_4]

    def get_rk4_arrays(self):
        """Get the arrays needed by the 4th order Runge-Kutta pusher."""
        return self.__dr_arrays, self.__dpr_arrays

    def allocate_rk4_field_arrays(self):
        """Allocate field arrays needed by the 4th order Runge-Kutta pusher.

        In order to compute the derivatives of r and pr at the 3 subteps
        of the RK4 pusher, the field values at the location of the particles
        in these substeps are needed. This method allocates the arrays
        that will store these field values.
        """
        self.__a2_2 = np.zeros(self.n_part)
        self.__nabla_a2_2 = np.zeros(self.n_part)
        self.__b_t_0_2 = np.zeros(self.n_part)
        self.__b_t_2 = np.zeros(self.n_part)
        self.__psi_2 = np.zeros(self.n_part)
        self.__dr_psi_2 = np.zeros(self.n_part)
        self.__dxi_psi_2 = np.zeros(self.n_part)
        self.__a2_3 = np.zeros(self.n_part)
        self.__nabla_a2_3 = np.zeros(self.n_part)
        self.__b_t_0_3 = np.zeros(self.n_part)
        self.__b_t_3 = np.zeros(self.n_part)
        self.__psi_3 = np.zeros(self.n_part)
        self.__dr_psi_3 = np.zeros(self.n_part)
        self.__dxi_psi_3 = np.zeros(self.n_part)
        self.__a2_4 = np.zeros(self.n_part)
        self.__nabla_a2_4 = np.zeros(self.n_part)
        self.__b_t_0_4 = np.zeros(self.n_part)
        self.__b_t_4 = np.zeros(self.n_part)
        self.__psi_4 = np.zeros(self.n_part)
        self.__dr_psi_4 = np.zeros(self.n_part)
        self.__dxi_psi_4 = np.zeros(self.n_part)
        self.__rk4_flds = [
            [
                self.__a2,
                self.__nabla_a2,
                self.__b_t_0,
                self.__b_t,
                self.__psi,
                self.__dr_psi,
                self.__dxi_psi,
            ],
            [
                self.__a2_2,
                self.__nabla_a2_2,
                self.__b_t_0_2,
                self.__b_t_2,
                self.__psi_2,
                self.__dr_psi_2,
                self.__dxi_psi_2,
            ],
            [
                self.__a2_3,
                self.__nabla_a2_3,
                self.__b_t_0_3,
                self.__b_t_3,
                self.__psi_3,
                self.__dr_psi_3,
                self.__dxi_psi_3,
            ],
            [
                self.__a2_4,
                self.__nabla_a2_4,
                self.__b_t_0_4,
                self.__b_t_4,
                self.__psi_4,
                self.__dr_psi_4,
                self.__dxi_psi_4,
            ],
        ]

    def get_rk4_field_arrays(self, i):
        """Get field arrays for the four substeps of the RK4 pusher."""
        return self.__rk4_flds[i]



"""Defines a method to compute the source terms from a particle bunch."""

import numpy as np
import scipy.constants as ct

from wake_t.particles.deposition import deposit_3d_distribution
from wake_t.utilities.numba import njit_serial


@njit_serial()
def calculate_bunch_source(q_bunch, n_r, n_xi, b_t_bunch):
    """
    Calculate the bunch source term (azimuthal magnetic field) from the
    charge deposited on the grid.

    Parameters
    ----------
    q_bunch : ndarray
        Array containing the bunch charge on the grid.
    n_r, n_xi : int
        Number of grid points along r and xi.
    r_grid : float
        Radial coordinated of the grid points.
    b_t_bunch : ndarray
        A (nz+4, nr+4) array where the magnetic field will be stored.
    """
    for i in range(n_xi):
        cumsum = 0.0
        # Iterate until n_r + 2 to get bunch source also in guard cells.
        for j in range(n_r + 2):
            q_ij = q_bunch[2 + i, 2 + j]
            cumsum += q_ij
            # At each grid cell, calculate integral only until cell center by
            # assuming that half the charge is evenly distributed within the
            # cell (i.e., subtract half the charge)
            b_t_bunch[2 + i, 2 + j] = (cumsum - 0.5 * q_ij) * 1.0 / (0.5 + j)
        # At the first grid point along r, subtract an additional 1/4 of the
        # charge. This comes from assuming that the density has to be zero on
        # axis.
        b_t_bunch[2 + i, 2] -= 0.25 * q_bunch[2 + i, 2] * 2.0


@njit_serial()
def deposit_bunch_charge(
    x,
    y,
    z,
    q,
    n_p,
    n_r,
    n_xi,
    r_grid,
    xi_grid,
    dr,
    dxi,
    p_shape,
    q_bunch,
    r_min_deposit=0.0,
):
    """
    Deposit the charge of particle bunch in a 2D grid.

    Parameters
    ----------
    x, y, z, q : ndarray
        The coordinates and charge of the bunch particles.
    n_p : float
        The plasma density.
    n_r, n_xi : int
        Number of grid points along r and xi.
    r_grid, xi_grid : float
        Coordinates of the grid nodes.
    dr, dxi : float
        Grid spacing in r and xi.
    p_shape : str
        Particle shape. Used to deposit the charge on the grid.
    q_bunch : ndarray
        A (nz+4, nr+4) array where the charge will be deposited.
    r_min_deposit : float
        The minimum radial position that particles must have in order to
        deposit their weight on the grid.

    Returns
    -------
    bool
        Whether all particles managed to deposit on the grid.
    """
    n_part = x.shape[0]
    s_d = ct.c / np.sqrt(ct.e**2 * n_p / (ct.m_e * ct.epsilon_0))
    k = 1.0 / (2 * np.pi * ct.e * dr * dxi * s_d * n_p)
    w = np.empty(n_part)
    for i in range(n_part):
        w[i] = q[i] * k

    return deposit_3d_distribution(
        z,
        x,
        y,
        w,
        xi_grid[0],
        r_grid[0],
        n_xi,
        n_r,
        dxi,
        dr,
        q_bunch,
        p_shape=p_shape,
        use_ruyten=True,
        r_min_deposit=r_min_deposit,
    )



"""This module contains methods for generating particle distributions"""

from typing import Optional

import numpy as np
import scipy.constants as ct
from scipy.stats import truncnorm
import aptools.plasma_accel.general_equations as ge
from aptools.particle_distributions.read import read_distribution
from aptools.data_handling.utilities import get_available_species

from wake_t.particles.particle_bunch import ParticleBunch


def get_gaussian_bunch_from_twiss(
    en_x: float,
    en_y: float,
    a_x: float,
    a_y: float,
    b_x: float,
    b_y: float,
    ene: float,
    ene_sp: float,
    s_t: float,
    xi_c: float,
    q_tot: float,
    n_part: int,
    x_off: Optional[float] = 0,
    y_off: Optional[float] = 0,
    theta_x: Optional[float] = 0,
    theta_y: Optional[float] = 0,
    name: Optional[str] = None,
    q_species: Optional[float] = -ct.e,
    m_species: Optional[float] = ct.m_e,
) -> ParticleBunch:
    """
    Creates a 6D Gaussian particle bunch with the specified Twiss parameters.

    Parameters
    ----------
    en_x, en_y : float
        Normalized trace-space emittance in x and y (units of m*rad).
    a_x, a_y : float
        Alpha Twiss parameter in x and y.
    b_x, b_y : float
        Beta Twiss parameter in x and y (units of m).
    ene : float
        Mean bunch energy in non-dimensional units (beta*gamma).
    ene_sp : float
        Relative energy spread in %.
    s_t : float
        Bunch duration (standard deviation) in units of fs.
    xi_c : float
        Central bunch position in the xi in units of m.
    q_tot : float
        Total bunch charge in pC.
    n_part : int
        Total number of particles in the bunch.
    x_off, y_off : float
        Centroid offsets in x and y (units of m).
    theta_x, theta_y : float
        Pointing angle in x and y (in radians).
    name : str
        Name of the particle bunch.
    q_species, m_species : float
        Charge and mass of a single particle of the species represented
        by the macroparticles. For an electron bunch (default),
        ``q_species=-e`` and ``m_species=m_e``

    Returns
    -------
    ParticleBunch
        The generated particle bunch.

    """

    # Calculate necessary values
    n_part = int(n_part)
    ene_sp = ene_sp / 100
    ene_sp_abs = ene_sp * ene
    s_z = s_t * 1e-15 * ct.c
    em_x = en_x / ene
    em_y = en_y / ene
    g_x = (1 + a_x**2) / b_x
    g_y = (1 + a_y**2) / b_y
    s_x = np.sqrt(em_x * b_x)
    s_y = np.sqrt(em_y * b_y)
    s_xp = np.sqrt(em_x * g_x)
    s_yp = np.sqrt(em_y * g_y)
    p_x = -a_x * em_x / (s_x * s_xp)
    p_y = -a_y * em_y / (s_y * s_yp)
    p_x_off = theta_x * ene
    p_y_off = theta_y * ene
    q_tot = q_tot / 1e12
    # Create normalized gaussian distributions
    u_x = np.random.standard_normal(n_part)
    v_x = np.random.standard_normal(n_part)
    u_y = np.random.standard_normal(n_part)
    v_y = np.random.standard_normal(n_part)
    # Calculate transverse particle distributions
    x = s_x * u_x + x_off
    xp = s_xp * (p_x * u_x + np.sqrt(1 - np.square(p_x)) * v_x)
    y = s_y * u_y + y_off
    yp = s_yp * (p_y * u_y + np.sqrt(1 - np.square(p_y)) * v_y)
    # Create longitudinal distributions (truncated at -3 and 3 sigma in xi)
    xi = truncnorm.rvs(-3, 3, loc=xi_c, scale=s_z, size=n_part)
    pz = np.random.normal(ene, ene_sp_abs, n_part)
    # Change from slope to momentum and apply offset
    px = xp * pz + p_x_off
    py = yp * pz + p_y_off
    # Macroparticle weight.
    w = np.abs(np.ones(n_part) * q_tot / (n_part * q_species))
    return ParticleBunch(
        w, x, y, xi, px, py, pz, name=name, q_species=q_species, m_species=m_species
    )


def get_gaussian_bunch_from_size(
    en_x: float,
    en_y: float,
    s_x: float,
    s_y: float,
    ene: float,
    ene_sp: float,
    s_t: float,
    xi_c: float,
    q_tot: float,
    n_part: int,
    x_off: Optional[float] = 0,
    y_off: Optional[float] = 0,
    theta_x: Optional[float] = 0,
    theta_y: Optional[float] = 0,
    name: Optional[str] = None,
    q_species: Optional[float] = -ct.e,
    m_species: Optional[float] = ct.m_e,
) -> ParticleBunch:
    """
    Creates a Gaussian bunch with the specified emitance and spot size. It is
    assumed to be on its waist (alpha_x = alpha_y = 0)

    Parameters
    ----------
    en_x, en_y : float
        Normalized trace-space emittance in x and y (units of m*rad).
    s_x, s_y : float
        Bunch size (standard deviation) in x and y (units of m).
    ene : float
        Mean bunch energy in non-dimensional units (beta*gamma).
    ene_sp : float
        Relative energy spread in %.
    s_t : float
        Bunch duration (standard deviation) in units of fs.
    xi_c : float
        Central bunch position in the xi in units of m.
    q_tot : float
        Total bunch charge in pC.
    n_part : int
        Total number of particles in the bunch.
    x_off, y_off : float
        Centroid offsets in x and y (units of m).
    theta_x, theta_y : float
        Pointing angle in x and y (in radians).
    name : str
        Name of the particle bunch.
    q_species, m_species : float
        Charge and mass of a single particle of the species represented
        by the macroparticles. For an electron bunch (default),
        ``q_species=-e`` and ``m_species=m_e``

    Returns
    -------
    ParticleBunch
        The generated particle bunch.

    """
    b_x = s_x**2 * ene / en_x
    b_y = s_y**2 * ene / en_y
    return get_gaussian_bunch_from_twiss(
        en_x,
        en_y,
        0,
        0,
        b_x,
        b_y,
        ene,
        ene_sp,
        s_t,
        xi_c,
        q_tot,
        n_part,
        x_off,
        y_off,
        theta_x,
        theta_y,
        name=name,
        q_species=q_species,
        m_species=m_species,
    )


def get_matched_bunch(
    en_x: float,
    en_y: float,
    ene: float,
    ene_sp: float,
    s_t: float,
    xi_c: float,
    q_tot: float,
    n_part: int,
    x_off: Optional[float] = 0,
    y_off: Optional[float] = 0,
    theta_x: Optional[float] = 0,
    theta_y: Optional[float] = 0,
    n_p: Optional[float] = None,
    k_x: Optional[float] = None,
    name: Optional[str] = None,
    q_species: Optional[float] = -ct.e,
    m_species: Optional[float] = ct.m_e,
) -> ParticleBunch:
    """
    Creates a Gaussian bunch matched to the plasma focusing fields.

    Parameters
    ----------
    en_x, en_y : float
        Normalized trace-space emittance in x and y (units of m*rad).
    ene : float
        Mean bunch energy in non-dimensional units (beta*gamma).
    ene_sp : float
        Relative energy spread in %.
    s_t : float
        Bunch duration (standard deviation) in units of fs.
    xi_c : float
        Central bunch position in the xi in units of m.
    q_tot : float
        Total bunch charge in pC.
    n_part : int
        Total number of particles in the bunch.
    x_off, y_off : float
        Centroid offsets in x and y (units of m).
    theta_x, theta_y : float
        Pointing angle in x and y (in radians).
    n_p : double
        Plasma density in units of m^{-3}. This value is used to calculate the
        focusing fields in the plasma assuming blowout regime.
    k_x : int
        Focusing fields in the plasma in units of T/m. Has priority over n_p.
    name : str
        Name of the particle bunch.
    q_species, m_species : float
        Charge and mass of a single particle of the species represented
        by the macroparticles. For an electron bunch (default),
        ``q_species=-e`` and ``m_species=m_e``

    Returns
    -------
    ParticleBunch
        The generated particle bunch.

    """
    if n_p is not None:
        n_p *= 1e-6
    b_m = ge.matched_plasma_beta_function(ene, n_p, k_x)
    return get_gaussian_bunch_from_twiss(
        en_x,
        en_y,
        0,
        0,
        b_m,
        b_m,
        ene,
        ene_sp,
        s_t,
        xi_c,
        q_tot,
        n_part,
        x_off,
        y_off,
        theta_x,
        theta_y,
        name=name,
        q_species=q_species,
        m_species=m_species,
    )


def get_from_file(
    file_path: str,
    data_format: str,
    preserve_prop_dist: Optional[bool] = False,
    name: Optional[str] = None,
    species_name: Optional[str] = None,
    **kwargs,
) -> ParticleBunch:
    """Get particle bunch from file.

    Parameters
    ----------
    file_path : str
        Path to the file containing the particle distribution.
    data_format : str
        Internal format of the data.  Possible values
        are 'astra', 'csrtrack' and 'openpmd'.
    preserve_prop_dist : bool, optional
        Whether to preserve the average longitudinal position of the
        distribution in the `prop_distance` parameter of the `ParticleBunch`,
        by default False.
    name : str, optional
        Name to assign to the ParticleBunch, by default None. If None, when
        reading a particle species from an `openPMD` file, the `name` is set
        to the original name of the species. For other data formats, the
        default name `elec_bunch_i` will be assigned, where `i`>=0 is an
        integer that is increased for each unnamed `ParticleBunch` that is
        generated.
    species_name : str, optional
        Name of the particle species to be read from an `openpmd` file.
        Required only when more than one particle species is present in the
        file.

    Other Parameters
    ----------------
    **kwargs : dict
        Other parameters to be passed to
        `aptools.data_handling.reading.read_beam`.

    Returns
    -------
    ParticleBunch
        A ParticleBunch with the distribution from the specified file.
    """
    # If reading from an openPMD file, use the right `name` and `species_name`.
    if data_format == "openpmd":
        if species_name is None:
            available_species = get_available_species(file_path)
            n_species = len(available_species)
            if n_species == 0:
                raise ValueError("No particle species found in '{}'".format(file_path))
            elif n_species == 1:
                species_name = available_species[0]
            else:
                raise ValueError(
                    "More than one particle species is available in"
                    + "'{}'. ".format(file_path)
                    + "Please specify a `species_name`. "
                    + "Available species are: "
                    + str(available_species)
                )
        kwargs["species_name"] = species_name
        if name is None:
            name = species_name
    distribution = read_distribution(file_path, data_format, **kwargs)
    # Center in z.
    z_avg = np.average(distribution.z, weights=distribution.w)
    xi = distribution.z - z_avg
    # Create ParticleBunch
    bunch = ParticleBunch(
        w=distribution.w,
        x=distribution.x,
        y=distribution.y,
        xi=xi,
        px=distribution.px,
        py=distribution.py,
        pz=distribution.pz,
        q_species=distribution.q_species,
        m_species=distribution.m_species,
        name=name,
    )
    # Preserve `z_avg` as the initial propagation distance of the bunch.
    if preserve_prop_dist:
        bunch.prop_distance = z_avg
    return bunch



import numpy as np
import matplotlib.pyplot as plt
import scipy.constants as ct
from wake_t.utilities.bunch_generation import get_gaussian_bunch_from_size
from wake_t.particles.deposition import deposit_3d_distribution


def test_gaussian_beam(show=False):
    """
    This test checks the accuracy of the charge deposition of a Gaussian beam
    into a cylindrical grid.

    It tests:
        (a) that the desposited charge equals the original charge.
        (b) that the charge density in the grid agrees with
            the expected charge density of a Gaussian beam.
        (c) that the resulting azimuthal magnetic field from the charge density
            agrees with the analytical expectations for a Gaussian
            distribution.

    The test is performed 4 times by using a low and a high resolution, and
    both the `'linear'` and `'cubic'` deposition methods.
    """
    # Set seed fo reproducible results.
    np.random.seed(0)

    # Bunch parameters
    s_x = 1e-6
    en = 1e-6
    gamma = 200
    ene_sp = 1e-5
    s_t = 10
    q_tot = 100
    n_part = 1e7
    s_z = s_t * (1e-15 * ct.c)
    s_r = s_x

    # Generate bunch
    bunch = get_gaussian_bunch_from_size(
        en, en, s_x, s_x, gamma, ene_sp, s_t, 0, q_tot, n_part=n_part
    )

    # Resolutions and shapes to test
    n_r_test = [20, 100]
    n_z_test = [41, 101]
    p_shape_test = ["linear", "cubic"]

    for n_r, n_z in zip(n_r_test, n_z_test):
        for p_shape in p_shape_test:
            # Generate grid
            z_min = -10e-6
            z_max = 10e-6
            r_max = 10e-6
            dr = r_max / n_r
            dz = (z_max - z_min) / (n_z - 1)
            r_fld = np.linspace(dr / 2, r_max - dr / 2, n_r)
            z_fld = np.linspace(z_min, z_max, n_z)
            rho_fld = np.zeros((n_z + 4, n_r + 4))

            # Deposit beam distribution.
            deposit_3d_distribution(
                bunch.xi,
                bunch.x,
                bunch.y,
                bunch.q,
                z_fld[0],
                r_fld[0],
                n_z,
                n_r,
                dz,
                dr,
                rho_fld,
                p_shape,
                use_ruyten=True,
            )

            # Remove guard cells.
            rho_fld = rho_fld[2:-2, 2:-2]

            # Check charge conservation.
            rel_q_dev = (np.sum(bunch.q) - np.sum(rho_fld)) / np.sum(bunch.q)
            assert rel_q_dev < 1e-12

            # Calculate charge density.
            q_cell = np.arange(n_r)
            v_cell = np.pi * ((q_cell + 1) ** 2 - q_cell**2) * dr**2 * dz
            rho_fld /= v_cell

            # Analytical expectation at beam center.
            rho_ana_0 = rho_gaussian_beam(r_fld, 0.0, -q_tot, s_r, s_z)

            # Deposited density at beam center.
            i_center = int((n_z - 1) / 2)
            rho_grid_0 = rho_fld[i_center]

            # Normalization factor.
            norm_factor = q_tot * 1e-12 / ((2 * np.pi) ** (3 / 2) * s_r**2 * s_z)

            # Check charge density conservation.
            assert np.max(np.abs(rho_grid_0 - rho_ana_0)) / norm_factor < 0.09

            # Calculate azimuthal magnetic field (analytical and from grid)
            b_theta_ana_0 = b_theta_gaussian_beam(r_fld, 0.0, q_tot, s_r, s_z, gamma)
            b_theta_fld = calculate_b_theta(rho_fld, r_fld, dr)
            b_theta_grid_0 = b_theta_fld[i_center]

            # Relative deviation w.r.t. analytical.
            rel_b_dev = np.abs((b_theta_grid_0 - b_theta_ana_0) / b_theta_ana_0)

            # Check long range fields are accurate.
            assert rel_b_dev[-1] < 3e-3

            # Check reasonable deviation in general. At low res, ~10%
            # differences can occur close to the axis.
            assert np.max(rel_b_dev) < 13e-2

            if show:
                plt.figure(figsize=(8, 4))
                plt.subplot(121)
                plt.plot(r_fld, rho_ana_0)
                plt.plot(r_fld, rho_grid_0)

                plt.subplot(122)
                plt.plot(r_fld, b_theta_ana_0)
                plt.plot(r_fld, b_theta_grid_0)

                plt.show()


def rho_gaussian_beam(r, z, q_tot, s_r, s_z):
    """Analytical charge density distribution of Gaussian beam."""
    return (
        q_tot
        * 1e-12
        / ((2 * np.pi) ** (3 / 2) * s_r**2 * s_z)
        * np.exp(-(z**2) / (2 * s_z**2))
        * np.exp(-(r**2) / (2 * s_r**2))
    )


def b_theta_gaussian_beam(r, z, q_tot, s_r, s_z, gamma):
    """Analytical azimuthal magnetic field for a Gaussian beam."""
    beta = np.sqrt(1 - 1 / gamma**2)
    e_r = (
        1
        / ((2 * np.pi) ** (3 / 2) * ct.epsilon_0)
        * q_tot
        * 1e-12
        / s_z
        * np.exp(-(z**2) / (2 * s_z**2))
        * (1 - np.exp(-(r**2) / (2 * s_r**2)))
        / r
    )
    b_theta = beta / ct.c * e_r
    return b_theta


def calculate_b_theta(rho, r_fld, dr):
    """Calculate azimuthal magnetic field from rho."""
    subs = rho * r_fld / 2
    subs[:, 0] += rho[:, 0] * r_fld[0] / 4
    b_theta = -(
        (np.cumsum(rho * r_fld, axis=1) - subs)
        * dr
        / np.abs(r_fld)
        / (ct.c * ct.epsilon_0)
    )
    return b_theta


if __name__ == "__main__":
    test_gaussian_beam(show=True)



"""
This module contains the methods for depositing the weights of a particle
distribution on a regular grid.

The deposition method is based on an adaptation of the same algorithm that is
implemented in FBPIC (https://github.com/fbpic/fbpic).

"""

import math
import numpy as np

from wake_t.utilities.numba import njit_serial, prange


@njit_serial()
def deposit_3d_distribution(
    z,
    x,
    y,
    w,
    z_min,
    r_min,
    nz,
    nr,
    dz,
    dr,
    deposition_array,
    p_shape="cubic",
    use_ruyten=False,
    r_min_deposit=0.0,
):
    """
    Deposit the the weight of each particle of a 3D distribution into a 2D
    grid (cylindrical symmetry).

    Parameters
    ----------
    z, x, y : arrays
        Arrays containing the longitudinal and transverse coordinates of the
        particles.
    w : array
        Weight of the particles (any quantity) which will be deposited into
        the grid.
    z_min : float
        Position of the first field value along z.
    r_min : float
        Position of the first field value along r.
    nz, nr : int
        Number of grid cells (excluding guard cells) along the longitudinal
        and radial directions.
    dz, dr : float
        Grid step size along the longitudinal and radial direction.
    deposition_array : array
        The 2D array of size (nr+4, nz+4) (including two guard cells at each
        boundary) into which the weight will be deposited (will be
        modified within this function)
    p_shape : str
        Particle shape to be used. Possible values are 'linear' or 'cubic'.
    r_min_deposit : float
        The minimum radial position that particles must have in order to
        deposit their weight on the grid.

    Returns
    -------
    bool
        Whether all particles managed to deposit on the grid.

    """
    if p_shape == "linear":
        return deposit_3d_distribution_linear(
            z,
            x,
            y,
            w,
            z_min,
            r_min,
            nz,
            nr,
            dz,
            dr,
            deposition_array,
            use_ruyten,
            r_min_deposit,
        )
    elif p_shape == "cubic":
        return deposit_3d_distribution_cubic(
            z,
            x,
            y,
            w,
            z_min,
            r_min,
            nz,
            nr,
            dz,
            dr,
            deposition_array,
            use_ruyten,
            r_min_deposit,
        )
    else:
        raise ValueError(
            "Particle shape not recognized. Possible values are 'linear' or 'cubic'."
        )


@njit_serial
def deposit_3d_distribution_linear(
    z,
    x,
    y,
    q,
    z_min,
    r_min,
    nz,
    nr,
    dz,
    dr,
    deposition_array,
    use_ruyten=False,
    r_min_deposit=0.0,
):
    """Calculate charge distribution assuming linear particle shape."""

    # Precalculate particle shape coefficients needed to satisfy charge
    # density conservation during deposition (see work by W.M. Ruyten
    # https://doi.org/10.1006/jcph.1993.1070).
    if use_ruyten:
        # Calculate the nr + 1 coefficients, where the first one is applied
        # to the particles located below the first grid point along r.
        ruyten_coef = np.zeros(nr + 1)
        r_grid = (np.arange(nr) + 0.5) * dr  # Assumes cell-centered in r.
        cell_volume = np.pi * dz * ((r_grid + 0.5 * dr) ** 2 - (r_grid - 0.5 * dr) ** 2)
        cell_volume_norm = cell_volume / (2 * np.pi * dr**2 * dz)
        cell_number = np.arange(nr) + 1
        ruyten_coef[1:] = (
            6.0
            / cell_number
            * (np.cumsum(cell_volume_norm) - 0.5 * cell_number**2 - 1.0 / 24)
        )

    z_max = z_min + (nz - 1) * dz
    r_max = nr * dr

    all_deposited = True

    # Loop over particles.
    for i in prange(z.shape[0]):
        # Get particle components.
        x_i = x[i]
        y_i = y[i]
        z_i = z[i]
        w_i = q[i]

        # Calculate radius.
        r_i = math.sqrt(x_i**2 + y_i**2)

        # Deposit only if particle is within field boundaries.
        if z_i >= z_min and z_i <= z_max and r_i >= r_min_deposit and r_i <= r_max:
            # Positions of the particles in cell units.
            r_cell = (r_i - r_min) / dr
            z_cell = (z_i - z_min) / dz

            # Indices of lowest cell in which the particle will deposit charge.
            ir_cell = min(int(math.ceil(r_cell)) + 1, nr + 2)
            iz_cell = int(math.ceil(z_cell)) + 1

            # u_r: particle position wrt left neighbor gridpoint in r.
            if r_cell < 0:
                # Force all charge to be deposited above axis.
                u_r = 1.0
            else:
                u_r = r_cell - int(math.ceil(r_cell)) + 1

            # u_z: particle position wrt left neighbor gridpoint in z.
            if z_cell < 0:
                # Force all charge to be deposited above z_min.
                u_z = 1.0
            elif r_cell > nz - 1:
                # Force all charge to be deposited below z_max.
                u_z = 0.0
            else:
                u_z = z_cell - int(math.ceil(z_cell)) + 1

            # Precalculate quantities.
            zsl_0 = 1.0 - u_z
            zsl_1 = u_z
            rsl_0 = 1.0 - u_r
            rsl_1 = u_r

            if use_ruyten:
                # Get corresponding coefficient for corrected shape factor.
                ir = min(int(math.ceil(r_cell)), nr)
                rc = ruyten_coef[ir]
                # Apply correction.
                rsl_0 += rc * (1.0 - u_r) * u_r
                rsl_1 -= rc * (1.0 - u_r) * u_r

            # Add contribution of particle to charge distribution.
            deposition_array[iz_cell + 0, ir_cell + 0] += zsl_0 * rsl_0 * w_i
            deposition_array[iz_cell + 0, ir_cell + 1] += zsl_0 * rsl_1 * w_i
            deposition_array[iz_cell + 1, ir_cell + 0] += zsl_1 * rsl_0 * w_i
            deposition_array[iz_cell + 1, ir_cell + 1] += zsl_1 * rsl_1 * w_i

        else:
            all_deposited = False
    return all_deposited


@njit_serial
def deposit_3d_distribution_cubic(
    z,
    x,
    y,
    q,
    z_min,
    r_min,
    nz,
    nr,
    dz,
    dr,
    deposition_array,
    use_ruyten=False,
    r_min_deposit=0.0,
):
    """Calculate charge distribution assuming cubic particle shape."""

    # Precalculate particle shape coefficients needed to satisfy charge
    # density conservation during deposition (see work by W.M. Ruyten
    # https://doi.org/10.1006/jcph.1993.1070).
    if use_ruyten:
        # Calculate the nr + 1 coefficients, where the first one is applied
        # to the particles located below the first grid point along r.
        ruyten_coef = np.zeros(nr + 1)
        r_grid = (np.arange(nr) + 0.5) * dr  # Assumes cell-centered in r.
        cell_volume = np.pi * dz * ((r_grid + 0.5 * dr) ** 2 - (r_grid - 0.5 * dr) ** 2)
        cell_volume_norm = cell_volume / (2 * np.pi * dr**2 * dz)
        cell_number = np.arange(nr) + 1
        ruyten_coef[1:] = (
            6.0
            / cell_number
            * (np.cumsum(cell_volume_norm) - 0.5 * cell_number**2 - 0.125)
        )
        ruyten_coef[1] = 6.0 * (cell_volume_norm[0] - 0.5 - 239.0 / (15 * 2**7))

    z_max = z_min + (nz - 1) * dz
    r_max = nr * dr

    all_deposited = True

    # Loop over particles.
    for i in prange(z.shape[0]):
        # Get particle components.
        x_i = x[i]
        y_i = y[i]
        z_i = z[i]
        w_i = q[i]

        # Calculate radius.
        r_i = math.sqrt(x_i**2 + y_i**2)

        # Deposit only if particle is within field boundaries.
        if z_i >= z_min and z_i <= z_max and r_i >= r_min_deposit and r_i <= r_max:
            # Positions of the particle in cell units.
            r_cell = (r_i - r_min) / dr
            z_cell = (z_i - z_min) / dz

            # Indices of lowest cell in which the particle will deposit charge.
            ir_cell = min(int(math.ceil(r_cell)), nr + 2)
            iz_cell = int(math.ceil(z_cell))

            # Particle position wrt left neighbor gridpoint.
            u_z = z_cell - int(math.ceil(z_cell)) + 1
            u_r = r_cell - int(math.ceil(r_cell)) + 1

            # Precalculate quantities for shape coefficients.
            inv_6 = 1.0 / 6.0
            v_z = 1.0 - u_z
            v_r = 1.0 - u_r

            # Cubic particle shape coefficients in z and r.
            zsc_0 = inv_6 * v_z**3
            zsc_1 = inv_6 * (3.0 * u_z**3 - 6.0 * u_z**2 + 4.0)
            zsc_2 = inv_6 * (3.0 * v_z**3 - 6.0 * v_z**2 + 4.0)
            zsc_3 = inv_6 * u_z**3
            rsc_0 = inv_6 * v_r**3
            rsc_1 = inv_6 * (3.0 * u_r**3 - 6.0 * u_r**2 + 4.0)
            rsc_2 = inv_6 * (3.0 * v_r**3 - 6.0 * v_r**2 + 4.0)
            rsc_3 = inv_6 * u_r**3

            if use_ruyten:
                # Get corresponding coefficient for corrected shape factor.
                ir = min(int(math.ceil(r_cell)), nr)
                rc = ruyten_coef[ir]
                # Add correction.
                rsc_1 += rc * v_r * u_r
                rsc_2 -= rc * v_r * u_r

            # Force all charge to be deposited within boundaries.
            # Below axis:
            if r_cell <= 0.0:
                rsc_3 += rsc_0
                rsc_2 += rsc_1
                rsc_0 = 0.0
                rsc_1 = 0.0
            elif r_cell <= 1.0:
                rsc_1 += rsc_0
                rsc_0 = 0.0
            # Below z_min:
            if z_cell <= 0.0:
                zsc_3 += zsc_0
                zsc_2 += zsc_1
                zsc_0 = 0.0
                zsc_1 = 0.0
            elif z_cell <= 1.0:
                zsc_1 += zsc_0
                zsc_0 = 0.0
            # Above z_max:
            elif z_cell > nz - 1:
                zsc_0 += zsc_3
                zsc_1 += zsc_2
                zsc_2 = 0.0
                zsc_3 = 0.0
            elif z_cell > nz - 2:
                zsc_2 += zsc_3
                zsc_3 = 0.0

            # Add contribution of particle to charge distribution.
            deposition_array[iz_cell + 0, ir_cell + 0] += zsc_0 * rsc_0 * w_i
            deposition_array[iz_cell + 0, ir_cell + 1] += zsc_0 * rsc_1 * w_i
            deposition_array[iz_cell + 0, ir_cell + 2] += zsc_0 * rsc_2 * w_i
            deposition_array[iz_cell + 0, ir_cell + 3] += zsc_0 * rsc_3 * w_i
            deposition_array[iz_cell + 1, ir_cell + 0] += zsc_1 * rsc_0 * w_i
            deposition_array[iz_cell + 1, ir_cell + 1] += zsc_1 * rsc_1 * w_i
            deposition_array[iz_cell + 1, ir_cell + 2] += zsc_1 * rsc_2 * w_i
            deposition_array[iz_cell + 1, ir_cell + 3] += zsc_1 * rsc_3 * w_i
            deposition_array[iz_cell + 2, ir_cell + 0] += zsc_2 * rsc_0 * w_i
            deposition_array[iz_cell + 2, ir_cell + 1] += zsc_2 * rsc_1 * w_i
            deposition_array[iz_cell + 2, ir_cell + 2] += zsc_2 * rsc_2 * w_i
            deposition_array[iz_cell + 2, ir_cell + 3] += zsc_2 * rsc_3 * w_i
            deposition_array[iz_cell + 3, ir_cell + 0] += zsc_3 * rsc_0 * w_i
            deposition_array[iz_cell + 3, ir_cell + 1] += zsc_3 * rsc_1 * w_i
            deposition_array[iz_cell + 3, ir_cell + 2] += zsc_3 * rsc_2 * w_i
            deposition_array[iz_cell + 3, ir_cell + 3] += zsc_3 * rsc_3 * w_i

        else:
            all_deposited = False
    return all_deposited



"""
This module contains the OpenPMDDiagnostics class, which generates the
openPMD output.
"""

import os
from typing import Optional, List

import numpy as np
import scipy.constants as ct
from openpmd_api import (
    Series,
    Access,
    Dataset,
    Mesh_Record_Component,
    Unit_Dimension,
    Geometry,
)

from wake_t import __version__
from wake_t.particles.particle_bunch import ParticleBunch
from wake_t.fields.base import Field


SCALAR = Mesh_Record_Component.SCALAR


class OpenPMDDiagnostics:
    """
    Class in charge of creating and writing the particle and field
    diagnostics following the openPMD standard.

    Parameters
    ----------
    write_dir : str
        Directory to which the diagnostics will be written. By default
        this will be a 'diags' folder in the current working directory.
    """

    def __init__(self, write_dir: Optional[str] = None) -> None:
        if write_dir is None:
            self.write_dir = os.path.join(os.getcwd(), "diags")
        else:
            self.write_dir = os.path.abspath(write_dir)
        # Index of the data file to be written. Will be increased after
        # each time step is written.
        self._index_out = 0
        # Longitudinal position at which the current beamline element begins.
        # This is needed in order to keep track of the global z position in the
        # simulation (needed for `globalOffset` and `positionOffset`
        # attributes) since the beamline elements themselves are not aware of
        # each other.
        self._current_z_pos = 0.0

    def write_diagnostics(
        self,
        time: float,
        dt: float,
        species_list: Optional[List[ParticleBunch]] = [],
        fields: Optional[List[Field]] = [],
    ) -> None:
        """
        Write to disk the diagnostics of a certain time step.

        Parameters
        ----------
        time : float
            Simulation time at the current beamline element.
        dt : float
            Time step used in the current beamline element.
        species_list : list, optional
            List of particle species to be written to file.
        wakefield : list, optional
            List of Fields that should be written to file.

        """
        # Perform checks.
        self.check_species_names(species_list)

        # Create diagnostics folder if it doesn't exist already.
        if not os.path.exists(self.write_dir):
            os.makedirs(self.write_dir)

        # Create file and series.
        file_name = "data%08T.h5"
        file_path = os.path.join(self.write_dir, "hdf5", file_name)
        opmd_series = Series(file_path, Access.create)

        # Set basic attributes.
        opmd_series.set_software("Wake-T", __version__)
        opmd_series.meshes_path = "fields"
        opmd_series.particles_path = "particles"
        opmd_series.openPMD_extension = 1

        # Create current iteration and set time attributes.
        it = opmd_series.iterations[self._index_out]
        it.time = time + self._current_z_pos / ct.c
        it.dt = dt

        # Write particle diagnostics.
        for species in species_list:
            diag_data = species.get_openpmd_diagnostics_data(it.time)
            self._write_species(it, diag_data)

        # Write field diagnostics.
        for field in fields:
            f_data = field.get_openpmd_diagnostics_data(it.time)
            if f_data is not None:
                self._write_fields(it, f_data)
                # Some field models might also have their own species
                # (e.g. plasma particles).
                if "species" in f_data.keys():
                    for species in f_data["species"]:
                        s_data = f_data[species]
                        self._write_species(it, s_data)

        # Flush data and increase counter for next step.
        opmd_series.flush()
        self._index_out += 1

    def increase_z_pos(self, dist: float) -> None:
        """
        Increase the current z position along the beamline. This should be
        called after tracking in each beamline element is completed.

        Parameters
        ----------
        dist : float
            This distance should be the length of the beamline element in
            which tracking has just finalized.

        """
        self._current_z_pos += dist

    def _write_species(self, it, species_data):
        """Write all particle diagnostics of a given species."""
        # Create particles for this species.
        particles = it.particles[species_data["name"]]

        # TODO: evaluate adding ED-PIC attributes to particles output. This is
        # tricky because Wake-T is not a PIC code and in the TMElements there
        # is no particle shape/smoothing nor charge/current deposition.
        # Could these attributes be only added in the time steps in which they
        # are actually used?

        if species_data["geometry"] == "3d_cartesian":
            components = ["x", "y", "z"]
        if species_data["geometry"] == "rz":
            components = ["r", "z"]

        # Create records of position components.
        position_record = False
        for component in components:
            if component in species_data:
                array = np.ascontiguousarray(species_data[component])
                pos_off = species_data["z_off"] if component == "z" else 0.0
                dataset = Dataset(array.dtype, extent=array.shape)
                offset = Dataset(np.dtype("float64"), extent=[1])
                particles["position"][component].reset_dataset(dataset)
                particles["positionOffset"][component].reset_dataset(offset)
                particles["position"][component].store_chunk(array)
                particles["positionOffset"][component].make_constant(pos_off)
                position_record = True
        if position_record:
            particles["position"].unit_dimension = {Unit_Dimension.L: 1}
            particles["positionOffset"].unit_dimension = {Unit_Dimension.L: 1}
            particles["position"].set_attribute("macroWeighted", np.uint32(0))
            particles["positionOffset"].set_attribute("macroWeighted", np.uint32(0))
            particles["position"].set_attribute("weightingPower", 0.0)
            particles["positionOffset"].set_attribute("weightingPower", 0.0)

        # Create records of momentum components.
        momentum_record = False
        for component in components:
            pc = "p" + component
            if pc in species_data:
                array = np.ascontiguousarray(species_data[pc])
                dataset = Dataset(array.dtype, extent=array.shape)
                particles["momentum"][component].reset_dataset(dataset)
                particles["momentum"][component].store_chunk(array)
                momentum_record = True
        if momentum_record:
            particles["momentum"].unit_dimension = {
                Unit_Dimension.L: 1,
                Unit_Dimension.M: 1,
                Unit_Dimension.T: -1,
            }
            particles["momentum"].set_attribute("macroWeighted", np.uint32(0))
            particles["momentum"].set_attribute("weightingPower", 1.0)

        # Do the same with geometry-independent data.
        if "w" in species_data:
            w = np.ascontiguousarray(species_data["w"])
            d_w = Dataset(w.dtype, extent=w.shape)
            particles["weighting"][SCALAR].reset_dataset(d_w)
            particles["weighting"][SCALAR].store_chunk(w)
            particles["weighting"][SCALAR].set_attribute("macroWeighted", np.uint32(1))
            particles["weighting"][SCALAR].set_attribute("weightingPower", 1.0)
        if "r_to_x" in species_data:
            r_to_x = np.ascontiguousarray(species_data["r_to_x"])
            d_r_to_x = Dataset(r_to_x.dtype, extent=r_to_x.shape)
            particles["r_to_x"][SCALAR].reset_dataset(d_r_to_x)
            particles["r_to_x"][SCALAR].store_chunk(r_to_x)
            particles["r_to_x"][SCALAR].set_attribute("macroWeighted", np.uint32(0))
            particles["r_to_x"][SCALAR].set_attribute("weightingPower", 1.0)
        if "id" in species_data:
            ids = np.ascontiguousarray(species_data["id"])
            d_id = Dataset(ids.dtype, extent=ids.shape)
            particles["id"][SCALAR].reset_dataset(d_id)
            particles["id"][SCALAR].store_chunk(ids)
            particles["id"][SCALAR].set_attribute("macroWeighted", np.uint32(0))
            particles["id"][SCALAR].set_attribute("weightingPower", 1.0)
        q = species_data["q"]
        m = species_data["m"]
        d_q = Dataset(np.dtype("float64"), extent=[1])
        d_m = Dataset(np.dtype("float64"), extent=[1])
        particles["charge"][SCALAR].reset_dataset(d_q)
        particles["mass"][SCALAR].reset_dataset(d_m)
        particles["charge"][SCALAR].make_constant(q)
        particles["mass"][SCALAR].make_constant(m)
        particles["charge"].unit_dimension = {
            Unit_Dimension.T: 1,
            Unit_Dimension.I: 1,
        }
        particles["mass"].unit_dimension = {Unit_Dimension.M: 1}
        particles["charge"][SCALAR].set_attribute("macroWeighted", np.uint32(0))
        particles["mass"][SCALAR].set_attribute("macroWeighted", np.uint32(0))
        particles["charge"][SCALAR].set_attribute("weightingPower", 1.0)
        particles["mass"][SCALAR].set_attribute("weightingPower", 1.0)

    def _write_fields(self, it, wf_data):
        """Write wakefield diagnostics."""
        # Set common field attributes.
        it.meshes.set_attribute("fieldSolver", wf_data["field_solver"])
        it.meshes.set_attribute("fieldSolverParams", wf_data["field_solver_params"])
        it.meshes.set_attribute("fieldBoundary", wf_data["field_boundary"])
        it.meshes.set_attribute("fieldBoundaryParams", wf_data["field_boundary_params"])
        it.meshes.set_attribute("particleBoundary", wf_data["particle_boundary"])
        it.meshes.set_attribute(
            "particleBoundaryParams", wf_data["particle_boundary_params"]
        )
        it.meshes.set_attribute("currentSmoothing", wf_data["current_smoothing"])
        it.meshes.set_attribute("chargeCorrection", wf_data["charge_correction"])

        # Add diagnostics of each field.
        for field in wf_data["fields"]:
            fld = it.meshes[field]

            if "comps" in wf_data[field]:
                for comp in wf_data[field]["comps"]:
                    fld_comp = fld[comp]
                    fld_comp_array = wf_data[field]["comps"][comp]["array"]
                    # Add extra dimension to mimic thetaMode geometry (mode 0).
                    # Workaround needed for reading the data with
                    # openPMD-viewer until cylindrical geometry is properly
                    # defined in the standard.
                    fld_comp_array = np.expand_dims(fld_comp_array, axis=0)
                    d_fld_comp = Dataset(
                        fld_comp_array.dtype, extent=fld_comp_array.shape
                    )
                    fld_comp.reset_dataset(d_fld_comp)
                    fld_comp.store_chunk(fld_comp_array)
                    fld_comp.set_attribute(
                        "position", wf_data[field]["comps"][comp]["position"]
                    )
            else:
                fld_array = wf_data[field]["array"]
                # Add extra dimmension to mimic thetaMode geometry (mode 0).
                fld_array = np.expand_dims(fld_array, axis=0)
                d_fld = Dataset(fld_array.dtype, extent=fld_array.shape)
                fld[SCALAR].reset_dataset(d_fld)
                fld[SCALAR].store_chunk(fld_array)
                fld[SCALAR].set_attribute("position", wf_data[field]["position"])

            if field == "E":
                fld.unit_dimension = {
                    Unit_Dimension.L: 1,
                    Unit_Dimension.M: 1,
                    Unit_Dimension.T: -3,
                    Unit_Dimension.I: -1,
                }
            elif field == "B":
                fld.unit_dimension = {
                    Unit_Dimension.M: 1,
                    Unit_Dimension.T: -2,
                    Unit_Dimension.I: -1,
                }
            elif field in ["rho", "chi"]:
                fld.unit_dimension = {
                    Unit_Dimension.L: -3,
                    Unit_Dimension.T: 1,
                    Unit_Dimension.I: 1,
                }
            elif field == "a":
                fld.unit_dimension = {
                    Unit_Dimension.L: 1,
                    Unit_Dimension.M: 1,
                    Unit_Dimension.T: -2,
                    Unit_Dimension.I: -1,
                }

            # Set geometry to thetaMode until cylindrical geometry is
            # properly defined in the openPMD standard.
            fld.geometry = Geometry.thetaMode  # Geometry.cylindrical
            fld.set_attribute("fieldSmoothing", "none")
            for attr, val in wf_data[field]["attributes"].items():
                fld.set_attribute(attr, val)
            fld.axis_labels = wf_data[field]["grid"]["labels"]
            fld.grid_spacing = wf_data[field]["grid"]["spacing"]
            fld.grid_global_offset = wf_data[field]["grid"]["global_offset"]

    def check_species_names(
        self, species_list: Optional[List[ParticleBunch]] = []
    ) -> None:
        """Check that no species have duplicate names."""
        sp_names = []
        for species in species_list:
            name = species.name
            if name not in sp_names:
                sp_names.append(name)
            else:
                raise ValueError("Several species share same name {}.".format(name))
